#=========================================================================
# test_utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Provide utility methods for testing."""

import copy
from collections import deque

import hypothesis.strategies as st

from pymtl3.datatypes import Bits1, mk_bits
from pymtl3.dsl import OutPort
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.stdlib.test import TestVectorSimulator

from ...yosys import TranslationImportPass as YosysTransImportPass
from .. import TranslationImportPass as SVTransImportPass

#=========================================================================
# test utility functions
#=========================================================================

def trim( s ):
  string = []
  lines = s.split( '\n' )
  for line in lines:
    _line = line.split()
    _string = "".join( _line )
    if _string and not _string.startswith( '//' ):
      string.append( "".join( line.split() ) )
  return "\n".join( string )

def check_eq( s, t ):
  if isinstance( s, list ) and isinstance( t, list ):
    for _s, _t in zip( s, t ):
      assert trim(_s) == trim(_t)
  else:
    assert trim(s) == trim(t)

#=========================================================================
# Hypothesis strategies for testing
#=========================================================================

def flatten( _rtype ):
  if isinstance( _rtype, rt.Array ):
    n_dim = _rtype.get_dim_sizes()
    rtype = _rtype.get_sub_type()
  else:
    n_dim = []
    rtype = _rtype
  return n_dim, rtype

#-------------------------------------------------------------------------
# Generate initialization data for a signal
#-------------------------------------------------------------------------

def VectorInitData( dtype ):
  nbits = dtype.get_length()
  min_val, max_val = -(2**(nbits-1)), 2**(nbits-1)-1
  value = 0
  return mk_bits( nbits )( value )

def StructInitData( dtype ):
  data = dtype.get_class()()
  for field_name, field in dtype.get_all_properties().items():
    setattr( data, field_name, DataTypeInitData( field ) )
  return data

def PackedArrayInitData( n_dim, sub_dtype ):
  if not n_dim:
    return DataTypeInitData( sub_dtype )
  else:
    data = []
    for i in range(n_dim[0]):
      data += [ PackedArrayInitData( n_dim[1:], sub_dtype ) ]
    return data

def DataTypeInitData( dtype ):
  if isinstance( dtype, rdt.Vector ):
    return VectorInitData( dtype )
  elif isinstance( dtype, rdt.Struct ):
    return StructInitData( dtype )
  elif isinstance( dtype, rdt.PackedArray ):
    n_dim = dtype.get_dim_sizes()
    sub_dtype = dtype.get_sub_dtype()
    return PackedArrayInitData( n_dim, sub_dtype )
  else:
    assert False, f"unrecognized data type {sub_dtype}!"

def InPortInitData( id_, port ):
  return { id_ : DataTypeInitData( port.get_dtype() ) }

def InterfaceInitData( id_, ifc ):
  init = {}
  for prop_name, prop_rtype in ifc.get_all_properties_packed():
    if isinstance( prop_rtype, rt.Array ):
      n_dim = prop_rtype.get_dim_sizes()
      sub_type = prop_rtype.get_sub_type()
      if isinstance(sub_type, rt.Port) and sub_type.get_direction() != "input":
        continue
      init.update( ArrayInitData( id_+"."+prop_name, n_dim, sub_type ) )
    elif isinstance( prop_rtype, rt.Port ):
      if prop_rtype.get_direction() == "input":
        init.update( InPortInitData( id_+"."+prop_name, prop_rtype ) )
    elif isinstance( prop_rtype, rt.InterfaceView ):
      init.update( InterfaceInitData( id_+"."+prop_name, prop_rtype ) )
  return init

def ArrayInitData( id_, n_dim, subtype ):
  if not n_dim:
    if isinstance( subtype, rt.Port ):
      return InPortInitData( id_, subtype )
    else:
      return InterfaceInitData( id_, subtype )
  else:
    init = {}
    for i in range(n_dim[0]):
      init.update( ArrayInitData(f'{id_}[{i}]', n_dim[1:], subtype) )
    return init

#-------------------------------------------------------------------------
# Hypothesis input data strategies
#-------------------------------------------------------------------------

@st.composite
def VectorDataStrategy( draw, dtype ):
  nbits = dtype.get_length()
  min_val, max_val = -(2**(nbits-1)), 2**(nbits-1)-1
  value = draw( st.integers( min_val, max_val ) )
  return mk_bits( nbits )( value )

@st.composite
def StructDataStrategy( draw, dtype ):
  data = dtype.get_class()()
  for field_name, field in dtype.get_all_properties().items():
    setattr( data, field_name, draw( DataTypeDataStrategy( field ) ) )
  return data

@st.composite
def PackedArrayDataStrategy( draw, n_dim, sub_dtype ):
  if not n_dim:
    return draw( DataTypeDataStrategy( sub_dtype ) )
  else:
    data = []
    for i in range(n_dim[0]):
      data += [ draw(PackedArrayDataStrategy( n_dim[1:], sub_dtype )) ]
    return data

@st.composite
def DataTypeDataStrategy( draw, dtype ):
  if isinstance( dtype, rdt.Vector ):
    return draw( VectorDataStrategy( dtype ) )
  elif isinstance( dtype, rdt.Struct ):
    return draw( StructDataStrategy( dtype ) )
  elif isinstance( dtype, rdt.PackedArray ):
    n_dim = dtype.get_dim_sizes()
    sub_dtype = dtype.get_sub_dtype()
    return draw( PackedArrayDataStrategy( n_dim, sub_dtype ) )
  else:
    assert False, f"unrecognized data type {sub_dtype}!"

@st.composite
def InPortDataStrategy( draw, id_, port ):
  return { id_ : draw(DataTypeDataStrategy( port.get_dtype() )) }

@st.composite
def InterfaceDataStrategy( draw, id_, ifc ):
  data = {}
  for prop_name, prop_rtype in ifc.get_all_properties_packed():
    if isinstance( prop_rtype, rt.Array ):
      n_dim = prop_rtype.get_dim_sizes()
      sub_type = prop_rtype.get_sub_type()
      if isinstance(sub_type, rt.Port) and sub_type.get_direction() != "input":
        continue
      data.update(draw(ArrayDataStrategy(id_+"."+prop_name, n_dim, sub_type)))
    elif isinstance( prop_rtype, rt.Port ):
      if prop_rtype.get_direction() == "input":
        data.update(draw(InPortDataStrategy(id_+"."+prop_name, prop_rtype)))
    elif isinstance( prop_rtype, rt.InterfaceView ):
      data.update(draw(InterfaceDataStrategy(id_+"."+prop_name, prop_rtype)))
  return data

@st.composite
def ArrayDataStrategy( draw, id_, n_dim, subtype ):
  if not n_dim:
    if isinstance( subtype, rt.Port ):
      return draw(InPortDataStrategy( id_, subtype ))
    else:
      return draw(InterfaceDataStrategy( id_, subtype ))
  else:
    data = {}
    for i in range(n_dim[0]):
      data.update(draw(
        ArrayDataStrategy(f'{id_}[{i}]', n_dim[1:], subtype)))
    return data

@st.composite
def DataStrategy( draw, dut ):
  """Return a strategy that generates input vector for component `dut`."""
  max_cycles = 10

  ret = []
  dut.elaborate()
  rifc = rt.get_component_ifc_rtlir( dut )
  ports = rifc.get_ports_packed()
  ifcs = rifc.get_ifc_views_packed()

  # Add reset cycle at the beginning
  reset1, reset2 = {}, {}
  for id_, port in ports:
    if id_ == "clk":
      reset1.update( { id_ : Bits1(0) } )
      reset2.update( { id_ : Bits1(1) } )
    elif id_ == "reset":
      reset1.update( { id_ : Bits1(1) } )
      reset2.update( { id_ : Bits1(1) } )
    else:
      n_dim, port_rtype = flatten( port )
      if port_rtype.get_direction() == "input":
        if n_dim:
          reset1.update( ArrayInitData( id_, n_dim, port_rtype ) )
          reset2.update( ArrayInitData( id_, n_dim, port_rtype ) )
        else:
          reset1.update( InPortInitData( id_, port_rtype ) )
          reset2.update( InPortInitData( id_, port_rtype ) )
  for id_, ifc in ifcs:
    n_dim, ifc_rtype = flatten( ifc )
    if n_dim:
      reset1.update( ArrayDataStrategy( id_, n_dim, ifc_rtype ) )
      reset2.update( ArrayDataStrategy( id_, n_dim, ifc_rtype ) )
    else:
      reset1.update( InterfaceInitData( id_, n_dim, ifc_rtype ) )
      reset2.update( InterfaceInitData( id_, n_dim, ifc_rtype ) )

  ret.append( reset1 )
  ret.append( reset2 )

  for i in range(max_cycles):
    data = {}
    for id_, port in ports:
      if id_ in [ "clk", "reset" ]:
        data.update( { id_ : Bits1(0) } )
      else:
        n_dim, port_rtype = flatten( port )
        if n_dim:
          if port_rtype.get_direction() == "input":
            data.update(draw( ArrayDataStrategy( id_, n_dim, port_rtype ) ))
        elif port_rtype.get_direction() == "input":
          data.update(draw( InPortDataStrategy( id_, port_rtype ) ))
    for id_, ifc in ifcs:
      n_dim, ifc_rtype = flatten( ifc )
      if n_dim:
        data.update(draw( ArrayDataStrategy( id_, n_dim, ifc_rtype ) ))
      else:
        data.update(draw( InterfaceDataStrategy( id_, ifc_rtype ) ))

    # Toggle clock signal
    toggle_data = {}
    for id_, signal in data.items():
      if id_ == "clk":
        toggle_data.update( { id_ : Bits1(1) } )
      else:
        toggle_data.update( { id_ : copy.deepcopy( signal ) } )

    ret.append( data )
    ret.append( toggle_data )

  return ret

#-------------------------------------------------------------------------
# closed_loop_component_input_test
#-------------------------------------------------------------------------

def closed_loop_component_input_test( dut, test_vector, tv_in, backend = "sverilog" ):

  # Filter to collect all output ports of a component
  def outport_filter( obj ):
    return isinstance( obj, OutPort )

  assert backend in [ "sverilog", "yosys" ], f"invalid backend {backend}!"

  dut.elaborate()
  reference_output = deque()
  all_output_ports = dut.get_local_object_filter( outport_filter )

  # Method to record reference outputs of the pure python component
  def ref_tv_out( model, test_vector ):
    dct = {}
    for out_port in all_output_ports:
      dct[ out_port ] = eval( "model." + out_port._dsl.my_name )
    reference_output.append( dct )

  # Method to compare the outputs of the imported model and the pure python one
  def tv_out( model, test_vector ):
    assert len(reference_output) > 0, \
      "Reference runs for fewer cycles than the imported model!"
    for out_port in all_output_ports:
      ref = reference_output[0][out_port]
      imp = eval( "model." + out_port._dsl.my_name )
      assert ref == imp, f"Value mismatch: reference: {ref}, imported: {imp}"
    reference_output.popleft()

  # First simulate the pure python component to see if it has sane behavior
  reference_sim = TestVectorSimulator( dut, test_vector, tv_in, ref_tv_out )
  reference_sim.run_test()
  dut.unlock_simulation()

  try:
    # If it simulates correctly, translate it and import it back
    dut.elaborate()
    if backend == "sverilog":
      dut.sverilog_translate_import = True
      imported_obj = SVTransImportPass()( dut )
    elif backend == "yosys":
      dut.yosys_translate_import = True
      imported_obj = YosysTransImportPass()( dut )
    # Run another vector simulator spin
    imported_sim = TestVectorSimulator( imported_obj, test_vector, tv_in, tv_out )
    imported_sim.run_test()
  finally:
    try:
      # Explicitly finalize the imported object because the shared lib may have
      # name collison
      imported_obj.finalize()
    except UnboundLocalError:
      # This test fails before the object is imported back
      pass

#-------------------------------------------------------------------------
# closed_loop_component_test
#-------------------------------------------------------------------------

def closed_loop_component_test( dut, data, backend = "sverilog" ):
  """Test the DUT with the given test_vector.

  User who wish to use this method should pass in the hypothesis data
  strategy instance as `data`. This method will reflect on the interfaces
  and ports of the given DUT and generate input vector.
  """
  # Method to feed data into the DUT
  def tv_in( model, test_vector ):
    for name, data in test_vector.items():
      # `setattr` fails to set the correct value of an array if indexed by
      # a subscript. We use `exec` here to make sure the value of elements
      # are assigned correctly.
      exec( "model." + name + " = data" )
  test_vector = data.draw( DataStrategy( dut ) )
  closed_loop_component_input_test( dut, test_vector, tv_in, backend )

#-------------------------------------------------------------------------
# closed_loop_test
#-------------------------------------------------------------------------

# TODO: A hypothesis test that works on generated test component AND
# generated input data.
def closed_loop_test():
  pass



#=========================================================================
# SVStructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 2 SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from ....testcases import (
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectConstStructAttrToOutComp,
    CaseConnectLiteralStructComp,
    CaseConnectNestedStructPackedArrayComp,
)
from ..SVStructuralTranslatorL2 import SVStructuralTranslatorL2


def run_test( case, m ):
  m.elaborate()
  SVStructuralTranslatorL2.is_sverilog_reserved = lambda s, x: x in sverilog_reserved
  tr = SVStructuralTranslatorL2( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  structs = tr.structural.decl_type_struct
  conns = tr.structural.connections[m]

  check_eq( ports, case.REF_PORT )
  check_eq( wires, case.REF_WIRE )
  check_eq( conns, case.REF_CONN )
  assert structs[-1][0] == case.REF_STRUCT[0]
  check_eq( structs[-1][1]['def'], case.REF_STRUCT[1] )

@pytest.mark.parametrize(
  'case', [
    CaseConnectConstStructAttrToOutComp,
    CaseConnectLiteralStructComp,
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectNestedStructPackedArrayComp,
  ]
)
def test_sverilog_structural_L2( case ):
  run_test( case, case.DUT() )



#=========================================================================
# RTLIRDataType.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jan 6, 2019
"""RTLIR instance data types and generation methods.

This file contains all RTLIR instance data types and methods to generate
these type objects. Each instance of the type class defined in this module
is a data type object or simply a data type. RTLIR instance type Signal
can be parameterized by the generated type objects.
"""
import inspect
from functools import reduce

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, is_bitstruct_class, is_bitstruct_inst

from ..errors import RTLIRConversionError
from ..util.utility import collect_objs


class BaseRTLIRDataType:
  """Base abstract RTLIR data type class."""
  def __ne__( s, other ):
    return not s.__eq__( other )

  def __hash__( s ):
    return hash(type(s))

class Vector( BaseRTLIRDataType ):
  """RTLIR data type class for vector type."""
  def __init__( s, nbits ):
    assert nbits > 0, 'vector bitwidth should be a positive integer!'
    s.nbits = nbits

  def get_length( s ):
    return s.nbits

  def __eq__( s, other ):
    return (isinstance(other, Vector) and s.nbits == other.nbits) or \
           (s.nbits == 1 and isinstance(other, Bool))

  def __hash__( s ):
    return hash((type(s), s.nbits))

  def __call__( s, obj ):
    """Return if type `obj` be cast into type `s`."""
    return isinstance( obj, ( Vector, Bool ) )

  def __str__( s ):
    return f'Vector{s.nbits}'

class Struct( BaseRTLIRDataType ):
  """RTLIR data type class for struct type."""
  def __init__( s, name, properties, cls = None ):
    # As of Python 3.7, dict always preserves insertion order
    assert len(properties) > 0, 'struct has no fields!'
    s.name = name
    s.properties = properties
    s.cls = cls
    if cls is not None:
      try:
        file_name = inspect.getsourcefile( cls )
        s.file_info = f"{file_name}"
        # line_no = inspect.getsourcelines( cls )[1]
        # s.file_info = f"{file_name}:{line_no}"
      except OSError:
        s.file_info = f"Dynamically generated class {cls.__name__}"
    else:
      s.file_info = "Not available"

  def __eq__( s, u ):
    return isinstance(u, Struct) and s.name == u.name

  def __hash__( s ):
    return hash((type(s), s.name))

  def get_name( s ):
    return s.name

  def get_file_info( s ):
    return s.file_info

  def get_class( s ):
    return s.cls

  def get_length( s ):
    return sum( d.get_length() for d in s.properties.values() )

  def has_property( s, p ):
    return p in s.properties

  def get_property( s, p ):
    return s.properties[ p ]

  def get_all_properties( s ):
    return s.properties

  def __call__( s, obj ):
    """Return if obj be cast into type `s`."""
    return s == obj

  def __str__( s ):
    return f'Struct {s.name}'

class Bool( BaseRTLIRDataType ):
  """RTLIR data type class for struct type.

  Bool data type cannot be instantiated explicitly. It can only appear as
  the result of comparisons.
  """
  def __init__( s ):
    pass

  def __eq__( s, other ):
    return isinstance(other, Bool) or \
           (isinstance(other, Vector) and other.nbits==1)

  def get_length( s ):
    return 1

  def __call__( s, obj ):
    """Return if obj be cast into type `s`."""
    return isinstance( obj, ( Bool, Vector ) )

  def __str__( s ):
    return 'Bool'

class PackedArray( BaseRTLIRDataType ):
  """RTLIR data type class for packed array type."""
  def __init__( s, dim_sizes, sub_dtype ):
    assert isinstance( sub_dtype, BaseRTLIRDataType ), \
      f'non-RTLIR data type {sub_dtype} as sub type of array.'
    assert not isinstance( sub_dtype, PackedArray ), \
      'nested PackedArray is not allowed!'
    assert len( dim_sizes ) >= 1, \
      'PackedArray dimension count should be greater than 0!'
    assert sum( dim_sizes ) > 0, 'PackedArray should have at least one element!'
    s.dim_sizes = dim_sizes
    s.sub_dtype = sub_dtype

  def __eq__( s, other ):
    if not isinstance(other, PackedArray): return False
    if len( s.dim_sizes ) != len( other.dim_sizes ): return False
    if not all(a == b for a, b in zip(s.dim_sizes, other.dim_sizes)):
      return False
    return s.sub_dtype == other.sub_dtype

  def __hash__( s ):
    return hash((type(s), tuple(s.dim_sizes), s.sub_dtype))

  def get_length( s ):
    return s.sub_dtype.get_length()*reduce( lambda p,x: p*x, s.dim_sizes, 1 )

  def get_next_dim_type( s ):
    if len( s.dim_sizes ) == 1:
      return s.sub_dtype
    return PackedArray( s.dim_sizes[1:], s.sub_dtype )

  def get_dim_sizes( s ):
    return s.dim_sizes

  def get_sub_dtype( s ):
    return s.sub_dtype

  def __call__( s, obj ):
    """Return if obj can be cast into type `s`."""
    return s == obj

  def __str__( s ):
    return f'PackedArray{s.dim_sizes} of {s.sub_dtype}'

def _get_rtlir_dtype_struct( obj ):

  # Vector field
  if isinstance( obj, Bits ):
    return Vector( obj.nbits )

  # PackedArray field
  elif isinstance( obj, list ):
    assert len( obj ) > 0, 'list length should be greater than 0!'
    ref_type = _get_rtlir_dtype_struct( obj[0] )
    assert all( _get_rtlir_dtype_struct(i) == ref_type for i in obj ), \
      f'all elements of array {obj} must have the same type {ref_type}!'
    dim_sizes = []
    while isinstance( obj, list ):
      assert len( obj ) > 0, 'list length should be greater than 0!'
      dim_sizes.append( len( obj ) )
      obj = obj[0]
    return PackedArray( dim_sizes, _get_rtlir_dtype_struct( obj ) )

  # Struct field
  elif is_bitstruct_inst( obj ):
    cls = obj.__class__

    properties = { name: _get_rtlir_dtype_struct( getattr(obj, name) )
                    for name in cls.__bitstruct_fields__.keys() }

    return Struct(cls.__name__, properties, cls)

  else:
    assert False, str(obj) + ' is not allowed as a field of struct!'

def get_rtlir_dtype( obj ):
  """Return the RTLIR data type of obj."""
  try:
    assert not isinstance(obj, list), \
      'array datatype object should be a field of some struct!'

    # Signals might be parameterized with different data types
    if isinstance( obj, ( dsl.Signal, dsl.Const ) ):
      Type = obj._dsl.Type
      assert isinstance( Type, type )

      # Vector data type
      if issubclass( Type, Bits ):
        return Vector( Type.nbits )

      # python int object
      elif Type is int:
        return Vector( 32 )

      # Struct data type
      elif is_bitstruct_class( Type ):
        try:
          return _get_rtlir_dtype_struct( Type() )
        except TypeError:
          assert False, \
            f'__init__() of supposed struct {Type.__name__} should take 0 argument ( you can \
            achieve this by adding default values to your arguments )!'

      else:
        assert False, f'cannot convert object of type {Type} into RTLIR!'

    # Python integer objects
    elif isinstance( obj, int ):
      # Following the Verilog bitwidth rule: number literals have 32 bit width
      # by default.
      return Vector( 32 )

    # PyMTL Bits objects
    elif isinstance( obj, Bits ):
      return Vector( obj.nbits )

    # PyMTL BitStruct objects
    elif is_bitstruct_inst( obj ):
      return _get_rtlir_dtype_struct( obj )

    else:
      assert False, 'cannot infer the data type of the given object!'
  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( obj, msg )



"""
========================================================================
PortCheck_test.py
========================================================================

Author : Shunning Jiang
Date   : Dec 25, 2017
"""
from pymtl3.datatypes import Bits32
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import InPort, OutPort, Wire
from pymtl3.dsl.errors import InvalidConnectionError, SignalTypeError

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  for i in range(10):
    A.tick()

def test_illegal_inport_write():

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort( Bits32 )

      @s.update
      def up_B_write():
        s.in_[1:10] = 10

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.b = B()

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid input port write SignalTypeError.")

def test_illegal_inport_deep_write():

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort( Bits32 )

      @s.update
      def up_B_print():
        print(s.in_)

  class BWrap( ComponentLevel3 ):
    def construct( s ):
      s.b = B()

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.b = BWrap()

      @s.update
      def up_write_b_in():
        s.b.b.in_[1:10] = 10

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid input port write SignalTypeError.")

def test_legal_inport_write():

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort( Bits32 )

      @s.update
      def up_B_print():
        print(s.in_)

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.b = B()

      @s.update
      def up_write_b_in():
        s.b.in_[1:10] = 10

  _test_model( Top )

def test_illegal_outport_write():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort( Bits32 )

      @s.update
      def up_A_read():
        print(s.out)

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()

      @s.update
      def up_write_a_out():
        s.a.out[1:10] = 10

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid output port write SignalTypeError.")

def test_illegal_outport_deep_write():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort( Bits32 )

      @s.update
      def up_A_read():
        print(s.out)

  class AWrap( ComponentLevel3 ):
    def construct( s ):
      s.a = A()

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = AWrap()

      @s.update
      def up_write_a_out():
        s.a.a.out[1:10] = 10

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid output port write SignalTypeError.")

def test_legal_outport_write():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort( Bits32 )

      @s.update
      def up_A_write():
        s.out[0:2] = 2

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()

      @s.update
      def up_read_a_out():
        print(s.a.out)

  _test_model( Top )

def test_illegal_wire_write():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire( Bits32 )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()

      @s.update
      def up_write_a_out():
        s.a.wire[1:10] = 10

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid wire write SignalTypeError.")

def test_illegal_wire_read():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire( Bits32 )
      @s.update
      def up_write_wire():
        s.wire[1:10] = 10

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()

      @s.update
      def up_read_a_out():
        print(s.a.wire[1:10])

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid wire write SignalTypeError.")

def test_legal_port_connect():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)

      @s.update
      def up_A_write():
        s.out = 123

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort(int)

      @s.update
      def up_B_read():
        print(s.in_)

  class OutWrap(ComponentLevel3):
    def construct( s ):
      s.out = OutPort(int)
      s.a = A()( out = s.out )

      @s.update
      def up_out_read():
        print(s.out)

  class InWrap(ComponentLevel3):
    def construct( s ):
      s.in_ = InPort(int)
      s.b = B()( in_ = s.in_ )

      @s.update
      def up_in_read():
        print(s.in_)

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.i = InWrap()
      s.o = OutWrap()
      connect( s.o.out, s.i.in_ )

  _test_model( Top )

def test_illegal_same_host():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      @s.update
      def up_A_write():
        s.out = 123

  class AWrap(ComponentLevel3):
    def construct( s ):
      s.out = OutPort(int) # Wire is the same
      s.a = A()( out = s.out )

      s.in_ = InPort(int)

      connect( s.out, s.in_ )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.awrap = AWrap()

  try:
    _test_model( Top )
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_illegal_rdhost_is_wrhost_parent():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      @s.update
      def up_A_write():
        s.out = 123

  class AWrap(ComponentLevel3):
    def construct( s ):
      s.out = InPort(int) # Should be OutPort
      s.a   = A()( out = s.out )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.awrap = AWrap()

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid wire type SignalTypeError.")

def test_illegal_wrhost_is_rdhost_parent():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      @s.update
      def up_A_write():
        s.out = 123

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = OutPort(int) # Should be InPort
      @s.update
      def up_B_read():
        print(s.in_)

  class BWrap(ComponentLevel3):
    def construct( s ):
      s.in_ = InPort(int)
      s.b   = B()( in_ = s.in_ )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()
      s.b = BWrap()
      connect( s.a.out, s.b.in_ )

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid wire type SignalTypeError.")

def test_illegal_hosts_same_parent():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      @s.update
      def up_A_write():
        s.out = 123

  class B( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort(int)
      @s.update
      def up_B_read():
        print(s.in_)

  class BWrap(ComponentLevel3):
    def construct( s ):
      s.in_ = OutPort(int) # Should be InPort
      s.b   = B()( in_ = s.in_ )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.a = A()
      s.b = BWrap()
      connect( s.a.out, s.b.in_ )

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown invalid wire type SignalTypeError.")

def test_illegal_hosts_too_far():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      @s.update
      def up_A_write():
        s.out = 123

  class AWrap( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(int)
      s.A = A()( out = s.out )

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(int)
      s.A = AWrap()

      connect( s.wire, s.A.A.out )

  try:
    _test_model( Top )
  except SignalTypeError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown hosts too far SignalTypeError.")



"""
========================================================================
DataStruct_test.py
========================================================================

Author : Shunning Jiang
Date   : Apr 16, 2018
"""
from pymtl3.datatypes import Bits16, Bits32, bitstruct
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import InPort, OutPort, Wire
from pymtl3.dsl.errors import InvalidFFAssignError, MultiWriterError, NoWriterError

from .sim_utils import simple_sim_pass


@bitstruct
class SomeMsg:
  a: Bits32
  b: Bits32

def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  for i in range(10):
    A.tick()

# All situations when we allow nested data struct:
#
# 1. WR A.b, RD A (need to recognize implicit constraint between A.b and A)
#    WR A.b, WR A (detect 2-writer conflict)
#
# 2. WR A, RD A.b (need to recognize implicit constraint between A.b and A)
#    WR A, WR A.b (detect 2-writer conflict)
#
# 3. WR A.b, A|=y, RD y (need to mark A as writer in net {A,y})
#    WR A.b, A|=y, WR y (detect 2-writer conflict)
#
# 4. WR A, A.b|=y, RD y (need to mark A.b as writer in net {A.b,y})
#    WR A, A.b|=y, WR y (detect 2-writer conflict)
#
# 5. WR x, x|=A.b, RD A (need to recognize implicit constraint between A.b and A because the generated connection block has A.b = x)
#    WR x, x|=A.b, WR A (detect 2-writer conflict)
#
# 6. WR x, x|=A, RD A.b (need to recognize implicit constraint between A.b and A because the generated connection block has A = x)
#    WR x, x|=A, WR A.b (detect 2-writer conflict)
#
# 7. WR x, x|=A.b, A|=y, RD y (need to mark A as writer and recognize the implicit constraint)
#    WR x, x|=A.b, A|=y, WR y (detect 2-writer conflict)
#
# 8. WR x, x|=A, A.b|=y, RD y (need to mark A.b as writer and recognize the implicit constraint)
#    WR x, x|=A, A.b|=y, WR y (detect 2-writer conflict)
#
# We fix A.b and see what can A.b's ancestors do:
#
# RD A.b
#  - WR A       (A (=) A.b)
#  - WR x, x|=A (A (=) A.b, THE SAME AS the first case thanks to the generated upblk)
#
# WR A.b
#  - RD A       (A (=) A.b)
#  - WR A       (detect 2-writer conflict)
#  - A|=x       (mark A as writer in net {A,x})
#  - A|=x, WR x (detect 2-writer conflict, THE SAME AS the second case thanks to the generated upblk)
#
# A.b|=x
#  - WR A       (mark A.b as writer in net {A.b,y})
#  - A|=y, WR y (mark A.b as writer and recognize the implicit constraint)
#
# A.b|=x, WR(x)
#  - RD A       (A (=) A.b)
#  - WR A       (detect 2-writer conflict)
#  - A|=y       (mark A as writer and recognize the implicit constraint)
#  - A|=y, WR y (detect 2-writer conflict)

# RD A.b - WR A
def test_rd_A_b_wr_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A():
        s.A = SomeMsg( 12, 123 )

      @s.update
      def up_rd_A_b():
        assert s.A.a == 12 and s.A.b == 123

  _test_model( Top )

# WR A.b - WR A
def test_wr_A_b_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A_b():
        s.A.b = Bits32( 123 )

      @s.update
      def up_wr_A():
        s.A = SomeMsg( 12, 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A.b - RD A
def test_wr_A_b_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A_b():
        s.A.b = 123

      @s.update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# WR A.b - RD A, RD A.b
def test_wr_A_b_rd_A_rd_A_b_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A_b():
        s.A.b = Bits32( 123 )

      @s.update
      def up_rd_A():
        z = s.A

      @s.update
      def up_rd_A_b():
        assert s.A.b == 123

  _test_model( Top )

# WR A - RD A.a, RD A.b
def test_wr_A_rd_fields_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A():
        s.A = SomeMsg( 12, 123 )

      @s.update
      def up_rd_A_a():
        assert s.A.a == 12

      @s.update
      def up_rd_A_b():
        assert s.A.b == 123

  _test_model( Top )

# WR A.b - RD A, RD A.a
def test_wr_A_b_rd_A_rd_A_a_cannot_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A_b():
        s.A.b = Bits32( 123 )

      @s.update
      def up_rd_A():
        z = s.A

      @s.update
      def up_rd_A_a():
        assert s.A.a == 12

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_A_b" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A" # only one constraint

# RD A.b - WR x, x|=A
def test_connect_rd_A_b_wr_x_conn_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( SomeMsg )
      s.A  = Wire( SomeMsg )

      connect( s.x, s.A )

      @s.update
      def up_wr_x():
        s.x = SomeMsg( 12, 123 )

      @s.update
      def up_rd_A_b():
        assert s.A.b == 123

  _test_model( Top )

# WR A.b - A|=x
def test_connect_wr_A_b_rd_x_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( SomeMsg )
      s.A  = Wire( SomeMsg )

      connect( s.x, s.A )

      @s.update
      def up_wr_A_b():
        s.A.b = Bits32( 123 )

  _test_model( Top )

# FIXME
# WR A.b - A|=x, WR x.b
# def test_connect_wr_A_b_wr_x_b_conn_A_conflict():

  # class Top( ComponentLevel3 ):
    # def construct( s ):

      # s.x  = Wire( SomeMsg )
      # s.A  = Wire( SomeMsg )

      # s.x |= s.A

      # @s.update
      # def up_wr_A_b():
        # s.A.b = Bits32( 123 )

      # @s.update
      # def up_wr_x_b():
        # s.x.a = 12

  # _test_model( Top )

# WR A.b - A|=x, WR x
def test_connect_wr_A_b_wr_x_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( SomeMsg )
      s.A  = Wire( SomeMsg )

      connect( s.x, s.A )

      @s.update
      def up_wr_A_b():
        s.A.b = Bits32( 123 )

      @s.update
      def up_wr_x():
        s.x = SomeMsg( 12, 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A.b|=x, WR x - RD A
def test_connect_wr_x_conn_A_b_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )

      connect( s.A.b, s.x )

      @s.update
      def up_wr_x():
        s.x = Bits32( 123 )

      @s.update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# A.b|=x, WR x - WR A
def test_connect_wr_x_conn_A_b_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )

      connect( s.A.b, s.x )

      @s.update
      def up_wr_x():
        s.x = Bits32( 123 )

      @s.update
      def up_wr_A():
        s.A = SomeMsg( 12, 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A.b|=x, RD x - WR A
def test_connect_rd_x_conn_A_b_wr_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )

      connect( s.A.b, s.x )

      @s.update
      def up_wr_A():
        s.A = SomeMsg( 12, 123 )

      @s.update
      def up_rd_x():
        z = s.x

  _test_model( Top )

# A.b|=x, WR x - A|=y, WR y
def test_connect_wr_x_conn_A_b_wr_y_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )
      s.y  = Wire( SomeMsg )

      connect( s.A.b, s.x )
      connect( s.A,   s.y )

      @s.update
      def up_wr_x():
        s.x = Bits32( 123 )

      @s.update
      def up_wr_y():
        s.y = SomeMsg( 12, 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A.b|=x, WR x - A|=y, RD y
def test_connect_wr_x_conn_A_b_rd_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )
      s.y  = Wire( SomeMsg )

      connect( s.A.b, s.x )
      connect( s.A,   s.y )

      @s.update
      def up_wr_x():
        s.x = Bits32( 123 )

      @s.update
      def up_rd_y():
        z = s.y

  _test_model( Top )

# A.b|=x, RD x - A|=y, WR y
def test_connect_rd_x_conn_A_b_wr_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( SomeMsg )
      s.y  = Wire( SomeMsg )

      connect( s.A.b, s.x )
      connect( s.A,   s.y )

      @s.update
      def up_rd_x():
        z = s.x

      @s.update
      def up_wr_y():
        s.y = SomeMsg( 12, 123 )

  _test_model( Top )

def test_iterative_find_nets():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.w  = Wire( SomeMsg )
      s.x  = Wire( SomeMsg )
      s.y  = Wire( SomeMsg )
      s.z  = Wire( SomeMsg )

      connect( s.w, s.x ) # net1
      connect( s.x.a, s.y.a ) # net2
      connect( s.y, s.z ) # net3

      @s.update
      def up_wr_s_w():
        s.w = SomeMsg( 12, 123 )

  _test_model( Top )

def test_deep_connections():

  class Msg1:
    def __init__( s, a=0, b=0 ):
      s.a = int( a )
      s.b = Bits32( b )

  class Msg2:
    def __init__( s, a=Msg1(), b=Msg1() ):
      s.p = a
      s.q = b

  class Msg3:
    def __init__( s, a=Msg1(), b=Msg2(), c=0 ):
      s.x = a
      s.y = b
      s.z = int( c )

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.A  = Wire( Msg3 )
      s.x  = Wire( Msg1 )
      s.y  = Wire( Msg2 )
      s.z  = Wire( Msg3 )

      s.w  = Wire( int )

      connect( s.A.y.p, s.x )
      connect( s.A.z,   s.w )
      connect( s.A,     s.z )

      @s.update
      def up_z():
        yy = s.z

      @s.update
      def up_rd_x():
        zz = s.x

      @s.update
      def up_wr_y():
        s.w = Msg2( 12, 123 )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

def test_struct_with_list_of_bits():

  class B:
    def __init__( s, foo=42 ):
      s.foo = [ Bits32( foo ) for _ in range(5) ]

  class A( ComponentLevel3 ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      # PyMTL mistakenly takes s.in_.foo[1] as a single bit!
      connect( s.out, s.in_.foo[1] )

  a = A()
  a.elaborate()

def test_nested_struct_2d_array_index():

  @bitstruct
  class C:
    bar: Bits16

  @bitstruct
  class B:
    foo: Bits32
    bar: [ [C]*5 ] * 5

  class A( ComponentLevel3 ):
    def construct( s ):
      s.struct = InPort( B )
      s.out    = OutPort( C )
      s.out2   = OutPort( Bits16 )
      connect( s.struct.bar[1][4], s.out )
      connect( s.struct.bar[1][4].bar, s.out2 )

      s.wire = Wire( B )
      @s.update_ff
      def ffs():
        s.wire.bar <<= 1

  a = A()
  try:
    a.elaborate()
  except InvalidFFAssignError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidFFAssignError.")

# TODO better error message?
def test_ff_cannot_write_to_struct_field():

  @bitstruct
  class C:
    bar: Bits16

  @bitstruct
  class B:
    foo: Bits32
    bar: [ [C]*5 ] * 5

  class A( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire( B )
      @s.update_ff
      def ffs():
        s.wire.bar <<= 1

  try:
    _test_model( A )
  except InvalidFFAssignError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidFFAssignError.")



#=========================================================================
# SVNameMangle_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 15, 2019
"""Test the SystemVerilog name mangling."""

from pymtl3.passes.backends.sverilog.import_.test.SVNameMangle_test import (
    test_interface,
    test_interface_array,
    test_nested_interface,
    test_nested_interface_port_array,
    test_nested_struct,
    test_packed_array_port_array,
    test_port_2d_array,
    test_port_array,
    test_port_single,
    test_struct_port_array,
    test_struct_port_single,
)
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..ImportPass import ImportPass


def local_do_test( m ):
  m.elaborate()
  rtype = rt.get_component_ifc_rtlir( m )
  ipass = ImportPass()
  result = ipass.gen_packed_ports( rtype )
  assert result == m._ref_ports_yosys



#=========================================================================
# BehavioralRTLIRImplGen.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jan 2, 2019
"""Generate behavioral RTLIR implementation.

This file generates (1) the implementation of the BehavioralRTLIR ASDL
defined in BehavioralRTLIR.asdl which should reside in the same
directory as this file and (2) the implementation of BehavioralRTLIR
visualization pass. The generated implementation is printed to
BehavioralRTLIR.py under the same directory. BehavioralRTLIR
visualization pass is printed to BehavioralRTLIRVisualizationPass.py
under the same directory.
"""

class constructor:
  """Class of constructors that create the behavioral RTLIR AST node types."""

  impl_template = \
"""
class {constr_name}( BaseBehavioralRTLIR ):
  {init}{eq_m}
"""

  viz_impl_template = \
"""
  def visit_{constr_name}( s, node ):
    s.cur += 1
    local_cur = s.cur
    table_body = '{table_body}'
    table_opt = s.gen_table_opt( node )
    label = (s.table_header + table_body + table_opt + s.table_trail){label_trail}
    {body}
"""

  rast_header_str = \
"""\
#=========================================================================
# BehavioralRTLIR.py
#=========================================================================
{}
""".format( '"""'+ 'Provide behavioral RTLIR AST node types.\n\n\
This file is automatically generated by BehavioralRTLIRImplGen.py.' + '\n"""' )

  rast_base_def_str = \
"""
class BaseBehavioralRTLIR:
  {base_rtlir_doc}
  def __eq__( s, other ):
    return type(s) is type(other)

  def __ne__( s, other ):
    return not s.__eq__( other )
""".format( base_rtlir_doc = '"""Base class for all behavioral RTLIR AST nodes."""' )

  rast_visitor_str = \
"""
class BehavioralRTLIRNodeVisitor:
  {visitor_doc}
  def visit( self, node, *args ):
    method = 'visit_' + node.__class__.__name__
    visitor = getattr( self, method, self.generic_visit )
    return visitor( node, *args )

  def generic_visit( self, node, *args ):
    for field, value in vars(node).items():
      if isinstance( value, list ):
        for item in value:
          if isinstance( item, BaseBehavioralRTLIR ):
            self.visit( item, *args )
      elif isinstance( value, BaseBehavioralRTLIR ):
        self.visit( value, *args )
""".format( visitor_doc = '"""Class for behavioral RTLIR AST visitors."""' )

  viz_header_str = \
"""\
#=========================================================================
# BehavioralRTLIRVisualizationPass.py
#=========================================================================
{}

import os

from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir.rtype.RTLIRType import BaseRTLIRType

from .BehavioralRTLIR import BehavioralRTLIRNodeVisitor
""".format( '"""Provide visualization for behavioral RTLIR AST.\n\n\
Visualize Behavioral RTLIR using Graphviz packeage. The output graph is in PDF\n\
format.  This file is automatically generated by BehavioralRTLIRImplGen.py.\n\
"""' )

  viz_class_def_str = \
"""
class BehavioralRTLIRVisualizationPass( BasePass ):
  def __call__( s, model ):
    visitor = BehavioralRTLIRVisualizationVisitor()

    for blk in model.get_update_blocks():
      visitor.init( blk.__name__ )
      visitor.visit( model._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )
      visitor.dump()

class BehavioralRTLIRVisualizationVisitor( BehavioralRTLIRNodeVisitor ):
  def __init__( s ):
    s.output = 'unamed'
    s.output_dir = 'rast-viz'
    s.table_header = '<<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0"> '
    s.table_trail = ' </TABLE>>'

  def init( s, name ):
    from graphviz import Digraph
    s.g = Digraph(
      comment = 'BehavioralRTLIR Visualization of ' + name,
      node_attr = { 'shape' : 'plaintext' }
    )
    s.blk_name = name
    s.cur = 0

  def get_str( s, obj ):
    return str(obj).replace('<', '&lt;').replace('>', '&gt;')

  def gen_table_opt( s, node ):
    ret = ''
    if isinstance( node.Type, BaseRTLIRType ):
      ret = ' <TR><TD COLSPAN="2">Type: ' + node.Type.__class__.__name__ + '</TD></TR>'
      for name, obj in vars(node.Type).items():
        obj_str = s.get_str( obj )
        if not isinstance( obj, dict ):
          ret += ' <TR><TD>' + name + '</TD><TD>' + obj_str + '</TD></TR>'
        else:
          ret += ' <TR><TD>' + name + '</TD><TD>{' + obj_str + '}</TD></TR>'
    return ret

  def dump( s ):
    if not os.path.exists( s.output_dir ):
      os.makedirs( s.output_dir )
    s.g.render( s.output_dir + os.sep + s.blk_name )
"""

  def __init__( s, name, type_list, field_list ):
    """Entry point to initialize one type of AST node.

    `name` is the name of this constructor and should be capitalized.
    `type_list` is the list of node types of all parameters. `field_list` is
    the list of names of all parameters. If the constructor does not have any
    parameters, both lists will be None."""
    assert name[0].isupper()

    isNone = ( type_list is None ) and ( field_list is None )
    isSameLen = ( not isNone ) and ( len( type_list ) == len( field_list ) )
    assert isNone or isSameLen

    s.name = name
    s.type_list = type_list
    s.field_list = field_list

  def impl_str( s ):
    """Return the implementation of this constructor as a Python class."""
    constr_name = s.name
    if s.type_list is None:
      init = ''
      single_term = ''
      not_single_term = ''
      check_equal = ''

    else:
      # Generate statements for checking sub fields
      eq = []
      single_term = ''
      not_single_term = ''
      for t, f in zip( s.type_list, s.field_list ):
        if s.is_sequence( t ):
          eq.append( 'for x, y in zip( s.{field}, other.{field} ):'.format( field = f ) )
          eq.append( '  if x != y:' )
          eq.append( '    return False' )
        else:
          single_term += ' or s.{field} != other.{field}'.format( field = f )
          not_single_term += ' and s.{field} == other.{field}'.format( field = f )

      init_args = []
      for Type, field in zip( s.type_list, s.field_list ):
        init_args.append( field if not s.is_optional( Type ) else f'{field} = None' )

      params_name = ', ' + ', '.join( init_args )
      params_assign = '\n    '.join(
        "s.{field} = {field}".format(field = x) for x in s.field_list
      )
      check_equal = '\n    '.join( eq )
      if check_equal:
        check_equal += '\n    '
      init = """\
def __init__( s{params_name} ):
    {params_assign}

  """.format( **locals() )

    eq_template = """\
if not isinstance(other, {constr_name}){single_term}:
      return False
    {check_equal}return True"""
    if check_equal == '':
      eq = 'return isinstance(other, {constr_name}){not_single_term}'.format( **locals() )
    else:
      eq = eq_template.format( **locals() )

    if check_equal == '' and not_single_term == '':
      if init == '':
        eq_m = 'pass'
      else:
        eq_m = ''
    else:
      eq_m = """\
def __eq__( s, other ):
    {eq}""".format( **locals() )

    return constructor.impl_template.format(
      constr_name = s.name,
      init = init,
      eq_m = eq_m
    )

  def viz_impl_str( s ):
    """Return the implementation of the visualization visitor of this
    constructor as a Python function."""
    body = []
    if s.type_list is None:
      # No parameter for this BehavioralRTLIR node.
      # Just creating a single vertex is enough.
      body.append( "s.g.node( str( s.cur ), label = label )" )
      table_body = '<TR><TD COLSPAN="2">{name}</TD></TR>'.format( name = s.name )
      body_str = '\n    '.join( body )
      body_str = body_str
      label_trail = ''

    else:
      # 1. Create a vertex corresponding to this BehavioralRTLIR node
      # 2. Add edges between this BehavioralRTLIR node and all child nodes
      body.append( "s.g.node( str( s.cur ), label = label )" )
      # The top string of vertex label
      table_body = '<TR><TD COLSPAN="2">{name}</TD></TR>'.format( name = s.name )
      # Templates for built-in fields
      built_in_str = \
        '<TR><TD>{type_name}</TD><TD>{{{value}}}</TD></TR>'
      built_in_trail = \
        '.format({built_in_trail_body})'
      built_in_trail_body = []
      # Template for user-defined fields
      customized_str = \
      "s.g.edge( str({s}), str({t}), label = '{edge_label}'{edge_label_trail} )"

      # Process each field of this BehavioralRTLIR node
      for t, f in zip( s.type_list, s.field_list ):
        if s.is_built_in( t ):
          # Add this built-in type to the label string
          # Assume built-in types will never have sequence modifier
          built_in = built_in_str.format( type_name = f, value = f )
          built_in_trail_body.append( f + '=s.get_str(node.' + f + ")" )
          table_body += ' ' + built_in
        else:
          # Add the user-defined field to the label string
          if s.is_sequence( t ):
            # A sequence of customized types
            indented = []
            indented.append('for i, f in enumerate(node.{field}):'.format(field = f))
            indented.append( customized_str.format(
              s = 'local_cur',
              t = 's.cur+1',
              edge_label = f + '[{idx}]',
              edge_label_trail = '.format(idx = i)'
            ) )
            indented.append( 's.visit( f )' )
            indented = [indented[0]] + ['  '+x for x in indented[1:]]
            indented = s.check_optional( indented, t, f )
            body = body + indented
          else:
            # A single customized type
            indented = []
            indented.append( customized_str.format(
              s = 'local_cur',
              t = 's.cur+1',
              edge_label = f,
              edge_label_trail = ''
            ) )
            indented.append( 's.visit( node.{field} )'.format( field = f ) )
            indented = s.check_optional( indented, t, f )
            body = body + indented
      if built_in_trail_body == []:
        label_trail = ''
      else:
        label_trail = built_in_trail.format(
          built_in_trail_body = ', '.join( built_in_trail_body ) )
      body_str = '\n    '.join( body )

    return constructor.viz_impl_template.format(
      constr_name = s.name,
      table_body = table_body,
      label_trail = label_trail,
      body = body_str
    )

  def is_built_in( s, type_name ):
    if type_name[-1] in ('*', '?'):
      type_name = type_name[:-1]
    return type_name in [ 'identifier', 'int', 'object', 'bool', 'string' ]

  def is_sequence( s, type_name ):
    return type_name[-1] == '*'

  def is_optional( s, type_name ):
    return type_name[-1] == '?'

  def check_optional( s, strs, Type, f ):
    ret = strs
    if s.is_optional( Type ):
      for i, string in enumerate(strs):
        ret[i] = '  ' + string
      ret = [ f'if node.{f} is not None:' ] + ret
    return ret

  def __eq__( s, other ):
    return isinstance(other, constructor) and s.name == other.name

  def __ne__( s, other ):
    return not s.__eq__( other )

def parse_constructor( constr_str ):
  """Return the constructor object corresponding to the given string."""
  if constr_str.find( '(' ) != -1:
    # has parameters
    bracket_idx = constr_str.index( '(' )
    bracket_end_idx = constr_str.index( ')' )
    constr_name = constr_str[ 0 : bracket_idx ]
    params = constr_str[ bracket_idx + 1 : bracket_end_idx ].strip()
    params = params.split( ',' )
    type_list = []
    field_list = []
    for param in params:
      param_lst = param.strip().split()
      type_list.append( param_lst[0] )
      field_list.append( param_lst[1] )
  else:
    # no parameters
    constr_name = constr_str
    type_list = None
    field_list = None

  return constructor( constr_name, type_list, field_list )

def get_type( type_name ):
  """Return the type of the given node type string (without *? modifier)."""
  if type_name[ -1 ] == '*' or type_name[ -1 ] == '?':
    return type_name[ : -1 ]
  return type_name

def get_constr( module_str, start, end ):
  """Extract and return the constructor string from module_str[start:end];
  also return the first position past the constructor string."""
  constr_start = start
  # Remove leading spaces
  while module_str[ constr_start ] == ' ':
    constr_start += 1

  if module_str.find( '(', start, end ) != -1:
    # this constructor has parameters
    bracket_idx = module_str.find( ')', start, end )
    constr = module_str[ constr_start : bracket_idx+1 ]
    l = bracket_idx + 1
  else:
    # this constructor has no parameters
    l = constr_start
    while l < end and module_str[ l ] != ' ':
      l += 1
    constr = module_str[ constr_start : l ]

  return constr, l

def implement_module( module_str ):
  """Return a string that implements all constructors in the given module
  string."""
  start = 0
  node_type = set()
  built_in_node_type = { 'identifier', 'int', 'string', 'bool', 'object' }
  constr_list = []
  # constr_list = set()

  impl_str = constructor.rast_header_str + constructor.rast_base_def_str
  viz_impl_str = constructor.viz_header_str + constructor.viz_class_def_str

  # parse one node type at a time
  while module_str.find( '=', start ) != -1:
    assign_idx = module_str.find( '=', start )
    node_type_name = module_str[ start : assign_idx ].strip()
    node_type.add( node_type_name )

    # find the boundary of this node type
    boundary = module_str.find( '=', assign_idx + 1 )
    if boundary == -1:
      boundary = len( module_str )

    constructor_start = assign_idx + 1
    # check if there are multiple constructors
    while module_str.find( '|', constructor_start ) != -1:
      constructor_end = module_str.find( '|', constructor_start )
      if constructor_end >= boundary: break

      # parse each possible constructor and move to the next
      constr_str, l = get_constr(
        module_str, constructor_start, constructor_end
      )

      constr = parse_constructor( constr_str )

      if not constr in constr_list:
        constr_list.append( constr )
      else:
        raise Exception( 'duplicated constructor!' )
      # constr_list.add( parse_constructor( constr_str ) )

      constructor_start = constructor_end + 1

    # one constructor remaining
    constr_str, l = get_constr( module_str, constructor_start, boundary )
    constr = parse_constructor( constr_str )
    if not constr in constr_list:
      constr_list.append( constr )
    else:
      raise Exception( 'duplicated constructor!' )
    # constr_list.add( parse_constructor( constr_str ) )

    start = l

  # sanity check
  for constr in constr_list:
    if not constr.type_list is None:
      for constr_type in constr.type_list:
        assert get_type( constr_type ) in node_type.union( built_in_node_type )

  # generate implementation
  for constr in constr_list:
    impl_str += constr.impl_str()
    viz_impl_str += constr.viz_impl_str()

  impl_str += constructor.rast_visitor_str

  with open( 'BehavioralRTLIR.py', 'w' ) as output:
    output.write( impl_str )

  with open( 'BehavioralRTLIRVisualizationPass.py', 'w' ) as output:
    output.write( viz_impl_str )

def extract_module( asdl_str ):
  """Return the module name and the module string of the given asdl
  string."""
  module_name_start = asdl_str.index( 'module' ) + len( 'module' )
  module_name_end = asdl_str.index( '{' )
  module_str_end = asdl_str.index( '}' )

  module_name = asdl_str[ module_name_start : module_name_end ].strip()
  module_str  = asdl_str[ module_name_end + 1 : module_str_end ].strip()

  return module_name, module_str

# This file should be run first to generate the correct implementation
# of BehavioralRTLIR.
if __name__ == '__main__':
  with open( 'BehavioralRTLIR.asdl', 'r') as asdl_file:
    asdl_str = ''
    for line in asdl_file:
      if line.strip().startswith( '--' ) or ( not line.strip() ):
        continue
      asdl_str += line.strip() + ' '

    # BehavioralRTLIR module is the first one in the file
    module_name, module_str = extract_module( asdl_str )
    assert module_name == 'BehavioralRTLIR'

    implement_module( module_str )



"""
======================================================================
Crossbar_test.py
======================================================================
"""
from pymtl3 import *
from pymtl3.stdlib.test import TestVectorSimulator

from .Crossbar import Crossbar


#-----------------------------------------------------------------------
# run_test_crossbar
#-----------------------------------------------------------------------
def run_test_crossbar( cls, args, test_vectors ):

  model = cls( *args )

  n, T = args
  Tsel = mk_bits( clog2( n ) )

  # Define functions mapping the test vector to ports in model

  def tv_in( model, test_vector ):
    n = len( model.in_ )

    for i in range(n):
      model.in_[i] = T(test_vector[i])
      model.sel[i] = Tsel(test_vector[n+i])

  def tv_out( model, test_vector ):
    n = len( model.in_ )

    for i in range(n):
      assert model.out[i] == T(test_vector[n*2+i])

  # Run the test

  sim = TestVectorSimulator( model, test_vectors, tv_in, tv_out )
  sim.run_test()

#-----------------------------------------------------------------------
# test_crossbar3
#-----------------------------------------------------------------------

def test_crossbar3():
  model = Crossbar( 3, Bits16 )
  run_test_crossbar( Crossbar, (3, Bits16), [
    [ 0xdead, 0xbeef, 0xcafe, 0, 1, 2, 0xdead, 0xbeef, 0xcafe ],
    [ 0xdead, 0xbeef, 0xcafe, 0, 2, 1, 0xdead, 0xcafe, 0xbeef ],
    [ 0xdead, 0xbeef, 0xcafe, 1, 2, 0, 0xbeef, 0xcafe, 0xdead ],
    [ 0xdead, 0xbeef, 0xcafe, 1, 0, 2, 0xbeef, 0xdead, 0xcafe ],
    [ 0xdead, 0xbeef, 0xcafe, 2, 1, 0, 0xcafe, 0xbeef, 0xdead ],
    [ 0xdead, 0xbeef, 0xcafe, 2, 0, 1, 0xcafe, 0xdead, 0xbeef ],
  ])



"""
==========================================================================
strategies.py
==========================================================================
Hypothesis strategies for built-in data types.

Author : Shunning Jiang, Yanghui Ou
  Date : Dec 9, 2019
"""
import hypothesis
from hypothesis import strategies as st

from .bits_import import Bits, mk_bits
from .bitstructs import is_bitstruct_class

#-------------------------------------------------------------------------
# strategies.bits
#-------------------------------------------------------------------------
# Return the SearchStrategy for Bits_nbits with the support to limit
# min/max value and setting signedness

def bits( nbits, signed=False, min_value=None, max_value=None ):
  BitsN = mk_bits( nbits )

  if (min_value is not None or max_value is not None) and signed:
    raise ValueError("bits strategy currently doesn't support setting "
                     "signedness and min/max value at the same time")

  if min_value is None:
    min_value = (-(2**(nbits-1))) if signed else 0
  if max_value is None:
    max_value = (2**(nbits-1)-1)  if signed else (2**nbits - 1)

  strat = st.booleans() if nbits == 1 else st.integers( min_value, max_value )

  @st.composite
  def strategy_bits( draw ):
    return BitsN( draw( strat ) )

  return strategy_bits() # RETURN A STRATEGY INSTEAD OF FUNCTION

#-------------------------------------------------------------------------
# strategies.bitslists
#-------------------------------------------------------------------------
# Return the SearchStrategy for a list of Bits with the support of
# dictionary based min/max value limit

def bitslists( types, limit_dict=None ):
  # Make sure limit_dict becomes a dict, not None
  limit_dict = limit_dict or {}
  if not isinstance( limit_dict, dict ):
    raise TypeError( f"bitlist '{types}' doesn't not take '{limit_dict}' " \
                      "to specify min/max limit. Here only a dictionary " \
                      "like { 0:range(1,2), 1:range(3,4) } is accepted. " )

  # We capture the strategies inside a list inside closure of the
  # strategy. For each element, we recursively compose a strategy based on
  # the type and the field limit.

  strats = [ _strategy_dispatch( type_, limit_dict.get( i, None ) )
              for i, type_ in enumerate(types) ]

  @st.composite
  def strategy_list( draw ):
    return [ draw(strat) for strat in strats ]

  return strategy_list() # RETURN A STRATEGY INSTEAD OF FUNCTION

#-------------------------------------------------------------------------
# strategies.bitstructs
#-------------------------------------------------------------------------
# Return the SearchStrategy for bitstruct type T with the support of
# dictionary-based min/max value limit

def bitstructs( T, limit_dict=None ):
  # Make sure limit_dict becomes a dict, not None
  limit_dict = limit_dict or {}
  if not isinstance( limit_dict, dict ):
    raise TypeError( f"bitstruct '{T}' doesn't not take '{limit_dict}' " \
                      "to specify min/max limit. Here only a dictionary " \
                      "like { 'x':range(1,2), 'y':range(3,4), 'z': { ... } } is accepted. " )

  # We capture the fields of T inside a list inside closure of the
  # strategy. For each field, we recursively compose a strategy based on
  # the type and the field limit.

  strats = [ _strategy_dispatch( type_, limit_dict.get( name, None ) )
              for name, type_ in T.__bitstruct_fields__.items() ]

  @st.composite
  def strategy_bitstruct( draw ):
    # Since strats already preserves the order of bitstruct fields,
    # we can directly asterisk the generatort to pass in as *args
    return T( * (draw(strat) for strat in strats) )

  return strategy_bitstruct() # RETURN A STRATEGY INSTEAD OF FUNCTION

# Dispatch to construct corresponding strategy based on given type
# The type can be a list of types, a Bits type, or a nested
# bitstruct. For nested bitstruct, we recursively call the bitstruct(T)
# function to construct the strategy
def _strategy_dispatch( T, limit ):

  # User-specified search strategy, early exit
  if isinstance( limit, st.SearchStrategy ):
    return limit

  # a list of types
  if isinstance( T, list ):
    return bitslists( T, limit )

  # nested bitstruct
  if is_bitstruct_class( T ):
    return bitstructs( T, limit )

  # bits field, "leaf node", directly use bits strategy
  assert issubclass( T, Bits )
  if limit is None:
    return bits( T.nbits )

  # bits field with range limit
  assert isinstance( limit, range ), f"We only accept range as min/max value specifier, not {type(limit)}"
  assert limit.step == 1, f"We only accept step=1 range, not {limit}."
  assert limit.start < limit.stop, f"We only accept start < stop range, not {limit}"

  return bits( T.nbits, False, limit.start, limit.stop-1 )



#=========================================================================
# DynamicSchedulePass.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : Apr 19, 2019

import os
from collections import deque

import py

from pymtl3.datatypes import Bits, is_bitstruct_class
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError

from .SimpleSchedulePass import SimpleSchedulePass, dump_dag
from .SimpleTickPass import SimpleTickPass


class DynamicSchedulePass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    if hasattr( top, "_sched" ):
      raise Exception("Some schedule pass has already been applied!")

    top._sched = PassMetadata()

    self.schedule_intra_cycle( top )

    # Reuse simple's ff and flip schedule
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

  def schedule_intra_cycle( self, top ):

    # Construct the intra-cycle graph based on normal update blocks

    V   = top._dag.final_upblks - top.get_all_update_ff()

    G   = { v: [] for v in V }
    G_T = { v: [] for v in V } # transpose graph

    E = set()
    for (u, v) in top._dag.all_constraints: # u -> v
      if u in V and v in V:
        G  [u].append( v )
        G_T[v].append( u )
        E.add( (u, v) )

    if 'MAMBA_DAG' in os.environ:
      dump_dag( top, V, E )

    # Compute SCC using Kosaraju's algorithm

    SCCs, G_new = kosaraju_scc( G, G_T )

    # Perform topological sort on SCCs

    InD = { i: 0 for i in range(len(SCCs)) }
    for u, vs in G_new.items():
      for v in vs:
        InD[ v ] += 1

    scc_pred = {}
    scc_schedule = []

    Q = deque( [ i for i in range(len(SCCs)) if not InD[i] ] )
    for x in Q:
      scc_pred[ x ] = None

    while Q:
      u = Q.pop()
      scc_schedule.append( u )
      for v in G_new[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )
          scc_pred[ v ] = u

    assert len(scc_schedule) == len(SCCs)

    #---------------------------------------------------------------------
    # Now we generate super blocks for each SCC and produce final schedule
    #---------------------------------------------------------------------

    constraint_objs = top._dag.constraint_objs

    # Put the graph schedule to _sched
    top._sched.update_schedule = schedule = []

    scc_id = 0
    for i in scc_schedule:
      scc = SCCs[i]
      if len(scc) == 1:
        schedule.append( list(scc)[0] )
      else:

        # For each non-trivial SCC, we need to figure out a intra-SCC
        # linear schedule that minimizes the time to re-execute this SCC
        # due to value changes. A bad schedule may inefficiently execute
        # the SCC for many times, each of which changes a few signals.
        # The current algorithm iteratively finds the "entry block" of
        # the SCC and expand its adjancent blocks. The implementation is
        # to first find the actual entry point, and then BFS to expand the
        # footprint until all nodes are visited.

        tmp_schedule = []
        Q = deque()

        if scc_pred[i] is None:
          # We start bfs from the block that has the least number of input
          # edges in the SCC
          InD = { v: 0 for v in scc }
          for (u, v) in E: # u -> v
            if u in scc and v in scc:
              InD[ v ] += 1
          Q.append( max(InD, key=InD.get) )

        else:
          # We start bfs with the blocks that are successors of the
          # predecessor scc in the previous SCC-level topological sort.
          pred = set( SCCs[ scc_pred[i] ] )
          # Sort by names for a fixed outcome
          for x in sorted( scc, key = lambda x: x.__name__ ):
            for v in G_T[x]: # find reversed edges point back to pred SCC
              if v in pred:
                Q.append( x )

        # Perform bfs to find a heuristic schedule
        visited = set(Q)
        while Q:
          u = Q.popleft()
          tmp_schedule.append( u )
          for v in G[u]:
            if v in scc and v not in visited:
              Q.append( v )
              visited.add( v )

        scc_id += 1
        variables = set()
        for (u, v) in E:
          # Collect all variables that triggers other blocks in the SCC
          if u in scc and v in scc:
            variables.update( constraint_objs[ (u, v) ] )

        if len(variables) == 0:
          raise Exception("There is a cyclic dependency without involving variables."
                          "Probably a loop that involves update_once:\n{}".format(", ".join( [ x.__name__ for x in scc] )))

        # generate a loop for scc
        # Shunning: we just simply loop over the whole SCC block
        # TODO performance optimizations using Mamba techniques within a SCC block

        def gen_wrapped_SCCblk( s, scc, src ):
          from pymtl3.dsl.errors import UpblkCyclicError

          # TODO mamba?
          scc_tick_func = SimpleTickPass.gen_tick_function( scc )
          namespace = {}
          namespace.update( locals() )

          exec(py.code.Source( src ).compile(), namespace)
          return namespace['generated_block']

        template = """
          from copy import deepcopy
          def wrapped_SCC_{0}():
            N = 0
            while True:
              N += 1
              if N > 100:
                raise UpblkCyclicError("Combinational loop detected at runtime in {{{3}}} after 100 iters!")
              {1}
              scc_tick_func()
              if {2}:
                break
            # print "SCC block{0} is executed", num_iters, "times"
          generated_block = wrapped_SCC_{0}
        """

        copy_srcs  = []
        check_srcs = []
        # print_srcs = []

        for j, var in enumerate(variables):
          if issubclass( var._dsl.Type, Bits ):
            copy_srcs.append( f"t{j} = {var}.value" )
          elif is_bitstruct_class( var._dsl.Type ):
            copy_srcs.append( f"t{j} = {var}.clone()" )
          else:
            copy_srcs.append( f"t{j} = deepcopy({var})" )

          check_srcs.append( f"{var} == t{j}" )
          # print_srcs.append( f"print '{var}', {var}, _____tmp_{j}" )

        scc_block_src = template.format( scc_id,
                                         "; ".join( copy_srcs ),
                                         " and ".join( check_srcs ),
                                         ", ".join( [ x.__name__ for x in scc] ) )
                                         # "; ".join( print_srcs ) )

        schedule.append( gen_wrapped_SCCblk( top, tmp_schedule, scc_block_src ) )

def kosaraju_scc( G, G_T ):

    #---------------------------------------------------------------------
    # Run Kosaraju's algorithm to shrink all strongly connected components
    # (SCCs) into super nodes
    #---------------------------------------------------------------------

    # First dfs on G to generate reverse post-order (RPO)
    # Shunning: we emulate the system stack to implement non-recursive
    # post-order DFS algorithm. At the beginning, I implemented a more
    # succinct recursive DFS but it turned out that a 1500-depth chain in
    # the graph will reach the CPython max recursion depth.
    # https://docs.python.org/3/library/sys.html#sys.getrecursionlimit

    PO = []

    vertices = list(G.keys())
    # random.shuffle(vertices)
    visited = set()

    # The commented algorithm loyally emulates the system stack by storing
    # the loop index in each stack element and push only one new element
    # to stack in every iteration. This is basically what recursive dfs
    # does.
    #
    # for u in vertices:
    #   if u not in visited:
    #     stack = [ (u, False) ]
    #     while stack:
    #       u, idx = stack.pop()
    #       visited.add( u )
    #       if idx == len(G[u]):
    #         PO.append( u )
    #       else:
    #         while idx < len(G[u]) and G[u][-idx] in visited:
    #           idx += 1
    #         if idx < len(G[u]):
    #           stack.append( (u, idx) )
    #           stack.append( (G[u][-idx], 0) )
    #         else:
    #           PO.append( u )

    # The following algorithm push all adjacent elements to the stack at
    # once and later check visited set to avoid redundant visit (instead
    # of checking visited set when pushing element to the stack). I added
    # a second_visit flag to add the node to post-order.

    for u in vertices:
      stack = [ (u, False) ]
      while stack:
        u, second_visit = stack.pop()

        if second_visit:
          PO.append( u )
        elif u not in visited:
          visited.add( u )
          stack.append( (u, True) )
          for v in reversed(G[u]):
            stack.append( (v, False) )

    RPO = PO[::-1]

    # Second bfs on G_T to generate SCCs

    SCCs  = []
    v_SCC = {}
    visited = set()

    for u in RPO:
      if u not in visited:
        visited.add( u )
        scc = set()
        SCCs.append( scc )
        Q = deque( [u] )
        scc.add( u )
        while Q:
          u = Q.popleft()
          v_SCC[u] = len(SCCs) - 1
          for v in G_T[u]:
            if v not in visited:
              visited.add( v )
              Q.append( v )
              scc.add( v )

    # Construct a new graph of SCCs

    G_new = { i: set() for i in range(len(SCCs)) }

    for u, vs in G.items():
      for v in vs: # u -> v
        scc_u, scc_v = v_SCC[u], v_SCC[v]
        if scc_u != scc_v and scc_v not in G_new[ scc_u ]:
          G_new[ scc_u ].add( scc_v )

    return SCCs, G_new



#=========================================================================
# ImportPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 14, 2019
"""Provide a pass that imports arbitrary SystemVerilog modules."""


from pymtl3.passes.backends.sverilog import ImportConfigs
from pymtl3.passes.backends.sverilog import ImportPass as SVerilogImportPass
from pymtl3.passes.backends.sverilog.errors import SVerilogImportError
from pymtl3.passes.backends.sverilog.util.utility import (
    get_component_unique_name,
    make_indent,
)
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import get_component_ifc_rtlir


class ImportPass( SVerilogImportPass ):

  #-----------------------------------------------------------------------
  # Backend-specific methods
  #-----------------------------------------------------------------------

  def get_backend_name( s ):
    return "yosys"

  def get_config( s, m ):
    return m.config_yosys_import

  def get_translation_namespace( s, m ):
    return m._pass_yosys_translation

  #-------------------------------------------------------------------------
  # Name mangling functions
  #-------------------------------------------------------------------------

  def mangle_vector( s, d, id_, dtype ):
    return [ ( id_, rt.Port( d, dtype ) ) ]

  def mangle_struct( s, d, id_, dtype ):
    ret = []
    for name, field in dtype.get_all_properties().items():
      ret += s.mangle_dtype( d, id_+"__"+name, field )
    return ret

  def mangle_packed_array( s, d, id_, dtype ):

    def _mangle_packed_array( d, id_, dtype, n_dim ):
      if not n_dim:
        return s.mangle_dtype( d, id_, dtype )
      else:
        ret = []
        for i in range( n_dim[0] ):
          ret += _mangle_packed_array( d, id_+"__"+str(i), dtype, n_dim[1:] )
        return ret

    n_dim, sub_dtype = dtype.get_dim_sizes(), dtype.get_sub_dtype()
    return _mangle_packed_array( d, id_, sub_dtype, n_dim )

  def mangle_dtype( s, d, id_, dtype ):
    if isinstance( dtype, rdt.Vector ):
      return s.mangle_vector( d, id_, dtype )
    elif isinstance( dtype, rdt.Struct ):
      return s.mangle_struct( d, id_, dtype )
    elif isinstance( dtype, rdt.PackedArray ):
      return s.mangle_packed_array( d, id_, dtype )

    raise TypeError(f"unrecognized data type {dtype}!")

  def mangle_port( s, id_, port, n_dim ):
    if not n_dim:
      return s.mangle_dtype( port.get_direction(), id_, port.get_dtype() )
    else:
      ret = []
      for i in range( n_dim[0] ):
        ret += s.mangle_port( id_+"__"+str(i), port, n_dim[1:] )
      return ret

  #-------------------------------------------------------------------------
  # Python signal connections generation functions
  #-------------------------------------------------------------------------

  def gen_vector_conns( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    _lhs, _rhs = s._verilator_name(lhs), s._verilator_name(rhs)
    ret = [f"connect( s.{_lhs}, s.mangled__{_rhs} )"]
    return ret, pos + nbits

  def gen_struct_conns( s, d, lhs, rhs, dtype, pos, symbols ):
    dtype_name = dtype.get_class().__name__
    upblk_name = lhs.replace('.', '_DOT_').replace('[', '_LBR_').replace(']', '_RBR_')
    ret = [
      "@s.update",
      "def " + upblk_name + "():",
    ]
    if d == "output":
      ret.append( f"  s.{lhs} = {dtype_name}()" )
    # Patch `dtype_name` into the symbol dictionary
    if dtype_name not in symbols:
      symbols[dtype_name] = dtype.get_class()
    body = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for name, field in all_properties:
      # Use upblk to generate assignment to a struct port
      _ret, pos = s._gen_dtype_conns( d, lhs+"."+name, rhs+"__"+name, field, pos )
      body += _ret
    return ret + body, pos

  def _gen_vector_conns( s, d, lhs, rhs, dtype, pos ):
    nbits = dtype.get_length()
    l, r = pos, pos+nbits
    _lhs, _rhs = s._verilator_name( lhs ), s._verilator_name( rhs )
    if d == "input":
      ret = [f"  s.mangled__{_rhs} = s.{_lhs}"]
    else:
      ret = [f"  s.{_lhs} = s.mangled__{_rhs}"]
    return ret, r

  def _gen_struct_conns( s, d, lhs, rhs, dtype, pos ):
    ret = []
    all_properties = reversed(list(dtype.get_all_properties().items()))
    for name, field in all_properties:
      _ret, pos = s._gen_dtype_conns(d, lhs+"."+name, rhs+"__"+name, field, pos)
      ret += _ret
    return ret, pos

  def _gen_packed_array_conns( s, d, lhs, rhs, dtype, n_dim, pos ):
    if not n_dim:
      return s._gen_dtype_conns( d, lhs, rhs, dtype, pos )
    else:
      ret = []
      for idx in range(n_dim[0]):
        _lhs = lhs + f"[{idx}]"
        _rhs = rhs + f"__{idx}"
        _ret, pos = \
          s._gen_packed_array_conns( d, _lhs, _rhs, dtype, n_dim[1:], pos )
        ret += _ret
      return ret, pos

  def gen_port_conns( s, id_py, id_v, port, n_dim, symbols ):
    if not n_dim:
      d = port.get_direction()
      dtype = port.get_dtype()
      nbits = dtype.get_length()
      ret, pos = s.gen_dtype_conns( d, id_py, id_v, dtype, 0, symbols )
      assert pos == nbits, \
        "internal error: {} wire length mismatch!".format( id_py )
      return ret
    else:
      ret = []
      for idx in range(n_dim[0]):
        _id_py = id_py + f"[{idx}]"
        _id_v = id_v + f"__{idx}"
        ret += s.gen_port_conns( _id_py, _id_v, port, n_dim[1:], symbols )
      return ret

  #-------------------------------------------------------------------------
  # PyMTL combinational input and output generation functions
  #-------------------------------------------------------------------------

  def gen_dtype_comb( s, lhs, rhs, dtype, is_input ):

    def _gen_packed_comb( lhs, rhs, dtype, n_dim, is_input ):
      if not n_dim:
        return s.gen_dtype_comb( lhs, rhs, dtype, is_input )
      else:
        ret = []
        for i in range( n_dim[0] ):
          if is_input:
            _lhs = f"{lhs}__{i}"
            _rhs = f"{rhs}[{i}]"
          else:
            _lhs = f"{lhs}[{i}]"
            _rhs = f"{rhs}__{i}"
          ret += _gen_packed_comb( _lhs, _rhs, dtype, n_dim[1:] )
        return ret

    if isinstance( dtype, rdt.Vector ):
      nbits = dtype.get_length()
      if is_input:
        return s._gen_ref_write( lhs, rhs, nbits )
      else:
        return s._gen_ref_read( lhs, rhs, nbits )
    elif isinstance( dtype, rdt.Struct ):
      ret = []
      for name, field in dtype.get_all_properties().items():
        if is_input:
          _lhs = f"{lhs}__{name}"
          _rhs = f"{rhs}.{name}"
        else:
          _lhs = f"{lhs}.{name}"
          _rhs = f"{rhs}__{name}"
        ret += s.gen_dtype_comb( _lhs, _rhs, field )
      return ret
    elif isinstance( dtype, rdt.PackedArray ):
      n_dim = dtype.get_dim_sizes()
      sub_dtype = dtype.get_sub_dtype()
      return _gen_packed_comb( lhs, rhs, sub_dtype, n_dim, is_input )
    else:
      assert False, f"unrecognized data type {dtype}!"

  def gen_port_array_input( s, lhs, rhs, dtype, n_dim ):
    if not n_dim:
      return s.gen_dtype_comb( lhs, rhs, dtype, is_input = True )
    else:
      ret = []
      for i in range( n_dim[0] ):
        _lhs = f"{lhs}__{i}"
        _rhs = f"{rhs}[{i}]"
        ret += s.gen_port_array_input( _lhs, _rhs, dtype, n_dim[1:] )
      return ret

  def gen_port_array_output( s, lhs, rhs, dtype, n_dim ):
    if not n_dim:
      return s.gen_dtype_comb( lhs, rhs, dtype, is_input = False )
    else:
      ret = []
      for i in range( n_dim[0] ):
        _lhs = f"{lhs}[{i}]"
        _rhs = f"{rhs}__{i}"
        ret += s.gen_port_array_output( _lhs, _rhs, dtype, n_dim[1:] )
      return ret

  #-------------------------------------------------------------------------
  # PyMTL wrapper constraint generation function
  #-------------------------------------------------------------------------

  def _gen_constraints( s, name, n_dim, rtype ):
    if not n_dim:
      return [f"U( seq_upblk ) < RD( s.mangled__{name} ),"]
    else:
      ret = []
      for i in range( n_dim[0] ):
        ret += s._gen_constraints( f"{name}__{i}", n_dim[1:], rtype )
      return ret



"""
========================================================================
Connectable.py
========================================================================
Wires, ports, and interfaces, all inherited from Connectable.

Author : Shunning Jiang
Date   : Apr 16, 2018
"""
import types
from collections import deque

from pymtl3.datatypes import Bits, Bits1, mk_bits

from .errors import InvalidConnectionError
from .NamedObject import DSLMetadata, NamedObject
from .Placeholder import Placeholder


class Connectable:
  # I've given up maintaining adjacency list or disjoint set locally since
  # we need to easily disconnect things

  # Public API

  def get_host_component( s ):
    try:
      return s._dsl.host
    except AttributeError:
      try:
        host = s
        while not host.is_component():
          host = host.get_parent_object() # go to the component
        s._dsl.host = host
        return s._dsl.host
      except AttributeError:
        raise NotElaboratedError()

  def __ifloordiv__( s, other ):
    # Currently this basically implements connect( s, other ), but to
    # avoid circular import, we replicate the implementation of connect.

    host, s_connectable, o_connectable = _connect_check( s, other, internal=False )

    # ifloordiv is the only entry point for connecting lambda
    if isinstance( other, types.LambdaType ):
      host._create_assign_lambda( s, other )
    else:
      host._connect_dispatch( s, other, s_connectable, o_connectable )

    return s

def _connect_check( o1, o2, internal ):
  """ Note that internal=False means we are just calling this API
      internally so that we don't connect other unconnectable fields by
      name in the interface."""

  o1_connectable = False
  o2_connectable = False
  top = None

  # Get access to the top level component by identifying a connectable

  if isinstance( o1, Connectable ):
    o1_connectable = True
    top = o1._dsl.elaborate_top

  if isinstance( o2, Connectable ):
    o2_connectable = True
    o2_top = o2._dsl.elaborate_top

    if o1_connectable: assert o2._dsl.elaborate_top is top
    else:              top = o2_top

  if not o1_connectable and not o2_connectable:
    if internal:  return None, False, False

    raise InvalidConnectionError("class {} and class {} are both not connectable.\n"
                                  "  (when connecting {} to {})" \
        .format( type(o1), type(o2), repr(o1), repr(o2)) )

  # Get the component from elaborate_stack

  try:
    host = top._dsl.elaborate_stack[-1]
  except AttributeError:
    raise InvalidConnectionError("Cannot call connect after elaboration.\n"
                                 "- Please use top.add_connection(...) API.")

  if isinstance( host, Placeholder ):
    raise InvalidPlaceholderError( "Cannot call connect "
          "in a placeholder component.".format( blk.__name__ ) )

  # Not sure if there is any case where we cannot get the top plus it's
  # not an internal connect call
  assert host is not None, "Please contact pymtl3 developer about this error."

  return host, o1_connectable, o2_connectable

# Checking if two slices/indices overlap
def _overlap( x, y ):
  if isinstance( x, int ):
    if isinstance( y, int ):  return x == y
    else:                     return y.start <= x < y.stop
  else: # x is slice
    if isinstance( y, int ):  return x.start <= y < x.stop
    else:
      if x.start <= y.start:  return y.start < x.stop
      else:                   return x.start < y.stop

# internal class for connecting signals and constants, not named object
class Const( Connectable ):
  def __init__( s, Type, v, parent ):
    s._dsl = DSLMetadata()
    s._dsl.Type = Type
    s._dsl.const = v
    s._dsl.parent_obj = parent

  def __repr__( s ):
    if s._dsl.Type is int:
      return f"int({s._dsl.const})"
    return f"{repr(s._dsl.const)}"

  def get_parent_object( s ):
    try:
      return s._dsl.parent_obj
    except AttributeError:
      raise NotElaboratedError()

  def is_component( s ):
    return False

  def is_signal( s ):
    return False

  def is_interface( s ):
    return False

class Signal( NamedObject, Connectable ):

  def __init__( s, Type=Bits1 ):
    assert isinstance( Type, type ), "Use actual type instead of instance!"
    s._dsl.Type = Type
    s._dsl.type_instance = None

    s._dsl.slice  = None # None -- not a slice of some wire by default
    s._dsl.slices = {}
    s._dsl.top_level_signal = s

    s._dsl.needs_double_buffer = False

  def inverse( s ):
    pass

  def __getattr__( s, name ):
    if name[0] == '_': # private variables directly exit here
      return super().__getattribute__( name )

    if name not in s.__dict__:
      # Shunning: we move this from __init__ to here for on-demand type
      #           checking when the __getattr__ is indeed used.

      if s._dsl.type_instance is None:
        # Yanghui: this would break if another Type indeed has an nbits
        #          attribute.
        # try:  Type.nbits
        # except AttributeError: # not Bits type

        # FIXME: check if Type is actually a type?
        Type = s._dsl.Type
        if not issubclass( Type, Bits ):
          s._dsl.type_instance = Type()

      if s._dsl.type_instance is None:
        raise AttributeError("{} is not a signals with struct type, and has no attribute '{}'".format( s, name ))
      else:
        obj = getattr( s._dsl.type_instance, name )

      # We handle three cases here:
      # 1. If the object is list, we recursively generate lists of signals
      # 2. If the object is Bits, we use the Bits type
      # 3. Otherwise we just go for obj.__class__
      # Note that BitsN is a type now. 2 and 3 are actually unified.

      # Use deque to ensure BFS to ensure a[0] is accessed before a[1]
      Q = deque([ (obj, [], s, False) ])

      while Q:
        u, indices, parent, parent_is_list = Q.popleft()
        cls = u.__class__

        if cls is list:
          x = []
          for i, v in enumerate( u ):
            Q.append( ( v, indices+[i], x, True ) )

        else:
          x = s.__class__( cls )
          sd = s._dsl
          xd = x._dsl

          xd.type_instance = u
          xd.parent_obj = s
          xd.top_level_signal = sd.top_level_signal
          xd.elaborate_top = sd.elaborate_top

          xd.my_name   = name + "".join([ f"[{y}]" for y in indices ])
          xd.full_name = f"{sd.full_name}.{xd.my_name}"

        if parent_is_list:
          parent.append( x )
        else:
          parent.__dict__[ name ] = x

    return s.__dict__[ name ]

  def __setitem__( s, idx, v ):
    pass # I have to override this to support a[0:1] |= b

  def __getitem__( s, idx ):
    assert issubclass( s._dsl.Type, Bits )
    # Turn index into a slice
    if isinstance( idx, int ):
      sl = slice( idx, idx+1 )
    elif isinstance( idx, slice ):
      sl = idx
    else: assert False, "What the hell?"

    sl_tuple = (sl.start, sl.stop)

    if sl_tuple not in s.__dict__:
      assert 0 <= sl.start < sl.stop <= s._dsl.Type.nbits
      x = s.__class__( mk_bits( sl.stop - sl.start) )
      sd = s._dsl
      xd = x._dsl
      xd.parent_obj = s
      xd.top_level_signal = sd.top_level_signal
      xd.elaborate_top = sd.elaborate_top

      sl_str = f"[{sl.start}:{sl.stop}]"

      xd.my_name   = f"{sd.my_name}{sl_str}"
      xd.full_name = f"{sd.full_name}{sl_str}"

      xd.slice       = sl
      s.__dict__[ sl_tuple ] = sd.slices[ sl_tuple ] = x

    return s.__dict__[ sl_tuple ]

  def default_value( s ):
    return s._dsl.Type()

  #-----------------------------------------------------------------------
  # Public APIs (only can be called after elaboration)
  #-----------------------------------------------------------------------

  def is_component( s ):
    return False

  def is_signal( s ):
    return True

  def is_input_value_port( s ):
    return False

  def is_output_value_port( s ):
    return False

  def is_wire( s ):
    return False

  def is_interface( s ):
    return False

  # Note: We currently define a leaf signal as int/Bits type signal, as
  #       opposed to BitStruct or normal Python object. A sliced signal is
  #       not a leaf signal. A non-leaf signal cannot be sliced or be a
  #       sliced signal.

  def is_leaf_signal( s ):
    return ( issubclass( s._dsl.Type, Bits ) and not s.is_sliced_signal() ) or \
           (s._dsl.Type is int)

  def get_leaf_signals( s ):
    if s.is_sliced_signal(): return []
    if s.is_leaf_signal():   return [ s ]

    leaf_signals = []
    def recursive_getattr( m, instance ):
      for x in instance.__dict__:
        signal = getattr( m, x )
        if signal.is_leaf_signal():
          leaf_signals.append( signal )
        else:
          recursive_getattr( signal, instance.__dict__[x] )

    # OK now it's not Bits or int, let's instantiate it if it's never
    # accessed
    if s._dsl.type_instance is None:
      s._dsl.type_instance = s._dsl.Type()
    recursive_getattr( s, s._dsl.type_instance )
    return leaf_signals

  def is_sliced_signal( s ):
    return not s._dsl.slice is None

  def is_top_level_signal( s ):
    return s._dsl.top_level_signal is s

  def get_top_level_signal( s ):
    top = s._dsl.top_level_signal
    return s if top is None else top

  def get_sibling_slices( s ):
    if s._dsl.slice:
      parent = s.get_parent_object()
      ret = list(parent._dsl.slices.values())
      ret.remove( s )
      return ret
    return []

  def slice_overlap( s, other ):
    assert other.get_parent_object() is s.get_parent_object(), \
      "You are only allowed to pass in a sibling signal."
    return _overlap( s._dsl.slice, other._dsl.slice )

# These three subtypes are for type checking purpose
class Wire( Signal ):
  def inverse( s ):
    return Wire( s._dsl.Type )

class InPort( Signal ):
  def inverse( s ):
    return OutPort( s._dsl.Type )
  def is_input_value_port( s ):
    return True

class OutPort( Signal ):
  def inverse( s ):
    return InPort( s._dsl.Type )
  def is_output_value_port( s ):
    return True

class Interface( NamedObject, Connectable ):

  def inverse( s ):
    s._dsl.inversed = True
    return s

  # Override
  # The same reason as __call__ connection. For s.x = A().inverse(),
  # inverse is executed before setattr, so we need to delay it ...

  def _construct( s ):
    if not s._dsl.constructed:
      s.construct( *s._dsl.args, **s._dsl.kwargs )

      inversed = False
      if hasattr( s._dsl, "inversed" ):
        inversed = s._dsl.inversed

      if inversed:
        for name, obj in s.__dict__.items():
          if name[0] != '_': # filter private variables
            if isinstance( obj, Signal ):
              setattr( s, name, obj.inverse() )
            else:
              setattr( s, name, obj )

      s._dsl.constructed = True

  # We move the connect functionality to Component
  # def connect()

  #-----------------------------------------------------------------------
  # Public APIs (only can be called after elaboration)
  #-----------------------------------------------------------------------

  def is_component( s ):
    return False

  def is_signal( s ):
    return False

  def is_interface( s ):
    return True

# CallerPort is connected an exterior method, called by the component's
# update block
# CalleePort exposes the method in the component to outside world

class MethodPort( NamedObject, Connectable ):

  def construct( self, *args, **kwargs ):
    raise NotImplementedError("You can only instantiate Caller/CalleePort.")

  def __call__( self, *args, **kwargs ):
    return self.method( *args, **kwargs )

  def is_component( s ):
    return False

  def is_signal( s ):
    return False

  def is_method_port( s ):
    return False

  def is_interface( s ):
    return False

  def in_non_blocking_interface( s ):
    return s._dsl.in_non_blocking_ifc

class CallerPort( MethodPort ):
  def construct( self, *, Type=None ): # keyword only!
    self.Type = Type
    self.method = None
    self._dsl.in_non_blocking_ifc = False

  def is_callee_port( s ):
    return False

  def is_caller_port( s ):
    return True

class CalleePort( MethodPort ):
  def construct( self, *, Type=None, method=None ): # keyword only!
    self.Type = Type
    self.method = method
    self._dsl.in_non_blocking_ifc = False

  def is_callee_port( s ):
    return True

  def is_caller_port( s ):
    return False

class CallIfcRTL( Interface ):

  def construct( s, *args, **kwargs ):
    raise NotImplementedError("You can only instantiate CallerIfcRTL/CalleeIfcRTL.")

  def __str__( s ):
    try:
      trace_len = s.trace_len
      trace_fmt = s.trace_fmt
    except AttributeError:
      trace_len = 0
      trace_fmt = ''

      if s.MsgType is not None:
        trace_len += len( f'{s.MsgType()}' ) + 2
        trace_fmt += f"({{s.msg}})"

      if s.RetType is not None:
        trace_len += 1 + len( f'{s.RetType()}' )
        trace_fmt += f"={{s.ret}}"

      if trace_len == 0:
        trace_len = 1
        trace_fmt = ' '

      s.trace_len = trace_len
      s.trace_fmt = trace_fmt

    if       s.en and not s.rdy:  return "X".ljust( trace_len ) # Not allowed
    elif not s.en and     s.rdy:  return " ".ljust( trace_len ) # Idle
    elif not s.en and not s.rdy:  return "#".ljust( trace_len ) # Stall
    return trace_fmt.format( **vars() ).ljust( trace_len )

class NonBlockingIfc( Interface ):
  def construct( s, *args, **kwargs ):
    raise NotImplementedError("You can only instantiate CallerIfcCL/CalleeIfcCL.")

  def __call__( s, *args, **kwargs ):
    return s.method( *args, **kwargs )

  def __str__( s ):
    return s._str_hook()

  def _str_hook( s ):
    return f"{s._dsl.my_name}"

class BlockingIfc( Interface ):
  def construct( s, *args, **kwargs ):
    raise NotImplementedError("You can only instantiate CallerIfcFL/CalleeIfcFL.")

  def __call__( s, *args, **kwargs ):
    return s.method( *args, **kwargs )

  def __str__( s ):
    return s._str_hook()

  def _str_hook( s ):
    return f"{s._dsl.my_name}"

#-------------------------------------------------------------------------
# First-class method-based interfaces
#-------------------------------------------------------------------------

class CalleeIfcRTL( CallIfcRTL ):

  def construct( s, *, en=None, rdy=None, MsgType=None, RetType=None ): # keyword only!
    s.MsgType = s.RetType = None

    if en is not None:
      s.en  = InPort ( Bits1 )

    if rdy is not None:
      s.rdy = OutPort( Bits1 )

    if MsgType is not None:
      s.msg = InPort ( MsgType )
      s.MsgType = MsgType

    if RetType is not None:
      s.ret = OutPort( RetType )
      s.RetType = RetType

class CallerIfcRTL( CallIfcRTL ):

  def construct( s, *, en=None, rdy=None, MsgType=None, RetType=None ): # keyword only!
    s.MsgType = s.RetType = None

    if en is not None:
      s.en  = OutPort( Bits1 )

    if rdy is not None:
      s.rdy = InPort( Bits1 )

    if MsgType is not None:
      s.msg = OutPort( MsgType )
      s.MsgType = MsgType

    if RetType is not None:
      s.ret = InPort( RetType )
      s.RetType = RetType

class CalleeIfcCL( NonBlockingIfc ):
  def construct( s, *, Type=None, method=None, rdy=None ): # keyword only!
    s.Type = Type
    s.method = CalleePort( Type=Type, method=method )
    s.rdy    = CalleePort( method=rdy )

    s.method._dsl.in_non_blocking_ifc = True
    s.rdy._dsl.in_non_blocking_ifc    = True

    s.method._dsl.is_rdy = False
    s.rdy._dsl.is_rdy    = True

class CallerIfcCL( NonBlockingIfc ):

  def construct( s, *, Type=None ): # keyword only!
    s.Type = Type

    s.method = CallerPort( Type=Type )
    s.rdy    = CallerPort()

    s.method._dsl.in_non_blocking_ifc = True
    s.rdy._dsl.in_non_blocking_ifc    = True

    s.method._dsl.is_rdy = False
    s.rdy._dsl.is_rdy    = True

class CalleeIfcFL( BlockingIfc ):

  def construct( s, *, Type=None, method=None ):
    s.Type   = Type
    s.method = CalleePort( method=method, Type=Type )

class CallerIfcFL( BlockingIfc ):

  def construct( s, *, Type=None ):
    s.Type   = Type
    s.method = CallerPort( Type=Type )



#=========================================================================
# DelayPipeCL_test.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : May 7, 2018

import pytest

from pymtl3 import *
from pymtl3.stdlib.test import TestSinkCL, TestSrcCL, mk_test_case_table

from .DelayPipeCL import DelayPipeDeqCL, DelayPipeSendCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, dut_class, src_msgs, sink_msgs, latency, src_lat, sink_lat ):

    # Messge type

    # Instantiate models

    s.src  = TestSrcCL( None, src_msgs, 0, src_lat )
    s.dut  = dut_class( latency )
    s.sink = TestSinkCL( None, sink_msgs, 0, sink_lat )

    # Connect

    connect( s.src.send,  s.dut.enq )

    if   dut_class is DelayPipeDeqCL:
      @s.update
      def up_adapt():
        if s.dut.deq.rdy() and s.sink.recv.rdy():
          s.sink.recv( s.dut.deq() )

    elif dut_class is DelayPipeSendCL:
      connect( s.dut.send, s.sink.recv )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace(s ):
    return s.src.line_trace() + " >>> " + s.dut.line_trace() + " >>> " + s.sink.line_trace()


def run_cl_sim( th, max_cycles=5000 ):

  # Create a simulator

  th.apply( SimulationPass() )

  # Reset model

  th.reset()

  # Run simulation

  ncycles = 0
  print("{:3}:{}".format( ncycles, th.line_trace() ))
  while not th.done() and ncycles < max_cycles:
    th.tick()
    ncycles += 1
    print("{:3}:{}".format( ncycles, th.line_trace() ))

  # Force a test failure if we timed out

  assert ncycles < max_cycles

  th.tick()
  th.tick()
  th.tick()

#-------------------------------------------------------------------------
# Test Case Table
#-------------------------------------------------------------------------

def basic_msgs():
  return [
    1, 1,
    2, 2,
    3, 3,
    4, 4,
    5, 5,
    6, 6,
    7, 7,
    8, 8,
    9, 9,
    10, 10,
  ]

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

@pytest.mark.parametrize( **mk_test_case_table([
  (                      "msg_func        lat   src_lat sink_lat  "),
  [ "basic",             basic_msgs,      0,    0,      0          ],
  [ "basic_lat1",        basic_msgs,      1,    0,      0          ],
  [ "basic_lat2",        basic_msgs,      2,    0,      0          ],
  [ "basic_lat3",        basic_msgs,      3,    0,      0          ],
  [ "basic_lat4",        basic_msgs,      4,    0,      0          ],
  [ "basic_lat10",       basic_msgs,      10,   0,      0          ],
  [ "basic_3_14",        basic_msgs,      0,    3,      14         ],
  [ "basic_lat1_3_14",   basic_msgs,      1,    3,      14         ],
  [ "basic_lat4_3_14",   basic_msgs,      4,    3,      14         ],
  [ "basic_lat10_3_14",  basic_msgs,      10,   3,      14         ],
]) )
def test_delay_pipe_deq( test_params, dump_vcd ):
  msgs = test_params.msg_func()
  run_cl_sim( TestHarness( DelayPipeDeqCL, msgs[::2], msgs[1::2],
                           test_params.lat,
                           test_params.src_lat, test_params.sink_lat ) )

@pytest.mark.parametrize( **mk_test_case_table([
  (                      "msg_func        lat   src_lat sink_lat  "),
  [ "basic",             basic_msgs,      0,    0,      0          ],
  [ "basic_lat1",        basic_msgs,      1,    0,      0          ],
  [ "basic_lat2",        basic_msgs,      2,    0,      0          ],
  [ "basic_lat3",        basic_msgs,      3,    0,      0          ],
  [ "basic_lat4",        basic_msgs,      4,    0,      0          ],
  [ "basic_lat10",       basic_msgs,      10,   0,      0          ],
  [ "basic_3_14",        basic_msgs,      0,    3,      14         ],
  [ "basic_lat1_3_14",   basic_msgs,      1,    3,      14         ],
  [ "basic_lat4_3_14",   basic_msgs,      4,    3,      14         ],
  [ "basic_lat10_3_14",  basic_msgs,      10,   3,      14         ],
]) )
def test_delay_pipe_send( test_params, dump_vcd ):
  msgs = test_params.msg_func()
  run_cl_sim( TestHarness( DelayPipeSendCL, msgs[::2], msgs[1::2],
                           test_params.lat,
                           test_params.src_lat, test_params.sink_lat ) )



from .EnqDeqIfc import DeqIfcFL, DeqIfcRTL, EnqIfcFL, EnqIfcRTL
from .GetGiveIfc import GetIfcFL, GetIfcRTL, GiveIfcFL, GiveIfcRTL
from .ifcs_utils import enrdy_to_str, valrdy_to_str
from .MemMsg import MemMsgType, mk_mem_msg, mk_mem_req_msg, mk_mem_resp_msg
from .SendRecvIfc import (
    RecvCL2SendRTL,
    RecvIfcFL,
    RecvIfcRTL,
    RecvRTL2SendCL,
    SendIfcFL,
    SendIfcRTL,
)
from .ValRdyIfc import InValRdyIfc, OutValRdyIfc
from .XcelMsg import XcelMsgType, mk_xcel_msg, mk_xcel_req_msg, mk_xcel_resp_msg



#=========================================================================
# errors.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Provide exception types for translation errors."""

import sys
import traceback


class RTLIRTranslationError( Exception ):
  """Exception type for errors happening during translation of RTLIR."""
  def __init__( self, obj, msg ):
    obj = str(obj)
    _, _, tb = sys.exc_info()
    traceback.print_tb(tb)
    tb_info = traceback.extract_tb(tb)
    fname, line, func, text = tb_info[-1]
    return super().__init__(
      f"\nIn file {fname}, Line {line}, Method {func}:"
      f"\nError trying to translate top component {obj}:\n- {msg}"
      f"\n  {text}" )



#=========================================================================
# BehavioralRTLIRTmpVar_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the temporary variable generation of behavioral RTLIR passes."""

from pymtl3.passes.rtlir.behavioral import (
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseBits32TmpWireComp,
    CaseScopeTmpWireOverwriteConflictComp,
    CaseStructTmpWireComp,
    CaseTmpWireOverwriteConflictComp,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )
  m.apply( BehavioralRTLIRVisualizationPass() )
  ref = m._rtlir_tmpvar_ref
  ns = m._pass_behavioral_rtlir_type_check

  for tvar_name in ref.keys():
    assert tvar_name in ns.rtlir_tmpvars
    assert ns.rtlir_tmpvars[tvar_name] == ref[tvar_name]

def test_tmp_wire( do_test ):
  a = CaseBits32TmpWireComp.DUT()
  a.elaborate()
  a._rtlir_tmpvar_ref = {('u', 'upblk') : rt.Wire(rdt.Vector(32))}
  do_test( a )

def test_tmp_wire_struct( do_test ):
  a = CaseStructTmpWireComp.DUT()
  a.elaborate()
  a._rtlir_tmpvar_ref = \
    {('u', 'upblk') : rt.Wire(rdt.Struct('Bits32Foo', {'foo':rdt.Vector(32)}))}
  do_test( a )

def test_tmp_wire_overwrite_conflict_type( do_test ):
  a = CaseTmpWireOverwriteConflictComp.DUT()
  a.elaborate()
  with expected_failure( PyMTLTypeError, "conflicting type" ):
    do_test( a )

def test_tmp_scope_conflict_type( do_test ):
  a = CaseScopeTmpWireOverwriteConflictComp.DUT()
  a.elaborate()
  with expected_failure( PyMTLTypeError, "conflicting type" ):
    do_test( a )



#=========================================================================
# BehavioralRTLIRTypeCheckL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L3 behavioral RTLIR type check pass."""

from collections import OrderedDict

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL2Pass import BehavioralRTLIRTypeCheckVisitorL2


class BehavioralRTLIRTypeCheckL3Pass( BasePass ):
  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    if not hasattr( m, '_pass_behavioral_rtlir_type_check' ):
      m._pass_behavioral_rtlir_type_check = PassMetadata()
    m._pass_behavioral_rtlir_type_check.rtlir_freevars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_tmpvars = OrderedDict()
    m._pass_behavioral_rtlir_type_check.rtlir_accessed = set()

    visitor = BehavioralRTLIRTypeCheckVisitorL3(
      m,
      m._pass_behavioral_rtlir_type_check.rtlir_freevars,
      m._pass_behavioral_rtlir_type_check.rtlir_accessed,
      m._pass_behavioral_rtlir_type_check.rtlir_tmpvars
    )

    for blk in m.get_update_block_order():
      visitor.enter( blk, m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] )

class BehavioralRTLIRTypeCheckVisitorL3( BehavioralRTLIRTypeCheckVisitorL2 ):
  def __init__( s, component, freevars, accessed, tmpvars ):
    super().__init__( component, freevars, accessed, tmpvars )
    s.type_expect[ 'Attribute' ] = {
      'value':( (rt.Component, rt.Signal),
        'the base of an attribute must be one of: component, signal!' )
    }

  def visit_Attribute( s, node ):
    if isinstance( node.value.Type, rt.Signal ):
      dtype = node.value.Type.get_dtype()
      if not isinstance( dtype, rdt.Struct ):
        raise PyMTLTypeError( s.blk, node.ast,
          'attribute base should be a struct signal!'
        )
      if not dtype.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{dtype.get_name()} does not have field {node.attr}!' )
      dtype = dtype.get_property( node.attr )
      if isinstance( node.value.Type, rt.Port ):
        rtype = rt.Port( node.value.Type.get_direction(), dtype )
      elif isinstance( node.value.Type, rt.Wire ):
        rtype = rt.Wire( dtype )
      elif isinstance( node.value.Type, rt.Const ):
        obj = node.value.Type.get_object()
        if obj is None:
          rtype = rt.Const( dtype )
        else:
          try:
            rtype = rt.Const( dtype, getattr( obj, node.attr ) )
          except AttributeError:
            rtype = rt.Const( dtype )
      else:
        raise PyMTLTypeError( s.blk, node.ast,
          f'unrecognized signal type {node.value.Type}!' )
      node.Type = rtype

    else:
      super().visit_Attribute( node )

  def visit_StructInst( s, node ):
    cls = node.struct

    dtype = rdt.get_rtlir_dtype( cls() )
    all_properties = dtype.get_all_properties()

    if len( all_properties ) != len( node.values ):
      raise PyMTLTypeError( s.blk, node.ast,
        f"BitStruct {cls.__name__} has {len(all_properties)} fields but only {len(node.values)} arguments are given!" )

    all_types = zip( node.values, list(all_properties.items()) )

    for idx, ( value, ( name, field ) ) in enumerate( all_types ):
      s.visit( value )
      # Expect each argument to be a signal
      if not isinstance( value.Type, rt.Signal ):
        raise PyMTLTypeError( s.blk, node.ast,
          f"argument #{idx} has type {value.Type} but not a signal!" )
      v_dtype = value.Type.get_dtype()
      # Expect each argument to have data type which corresponds to the field
      if v_dtype != field:
        raise PyMTLTypeError( s.blk, node.ast,
          f"Expected argument#{idx} ( field {name} ) to have type {field}, but got {v_dtype}." )

    node.Type = rt.Const( dtype )



#=========================================================================
# YosysStructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 4 yosys-SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from ....testcases import (
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectArraySubCompArrayStructIfcComp,
)
from ..YosysStructuralTranslatorL4 import YosysStructuralTranslatorL4


def run_test( case, m ):
  m.elaborate()
  YosysStructuralTranslatorL4.is_sverilog_reserved = lambda s, x: x in sverilog_reserved
  tr = YosysStructuralTranslatorL4( m )
  tr.clear( m )
  tr.translate_structural( m )

  comps = tr.structural.decl_subcomps[m]

  check_eq( comps["port_decls"],  case.REF_COMP_PORT )
  check_eq( comps["wire_decls"],  case.REF_COMP_WIRE )
  check_eq( comps["connections"], case.REF_COMP_CONN )

@pytest.mark.parametrize(
  'case', [
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseBits32ArrayConnectSubCompAttrComp,
  ]
)
def test_yosys_structural_L4( case ):
  run_test( case, case.DUT() )



#=========================================================================
# BehavioralRTLIRFreeVar_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the free variable generation of behavioral RTLIR passes."""

from pymtl3.datatypes import Bits32, bitstruct
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir.behavioral import (
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
)
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseBits32ClosureConstruct,
    CaseBits32ClosureGlobal,
    CaseStructClosureGlobal,
    pymtl_Bits_global_freevar,
)


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )
  ref = m._rtlir_freevar_ref
  ns = m._pass_behavioral_rtlir_type_check

  for fvar_name in ref.keys():
    assert fvar_name in ns.rtlir_freevars
    assert ns.rtlir_freevars[fvar_name] == ref[ fvar_name ]

def test_pymtl_Bits_closure_construct( do_test ):
  a = CaseBits32ClosureConstruct.DUT()
  a.elaborate()
  a._rtlir_freevar_ref = { 'foo' : a.fvar_ref }
  do_test( a )

def test_pymtl_Bits_global( do_test ):
  a = CaseBits32ClosureGlobal.DUT()
  a.elaborate()
  a._rtlir_freevar_ref = \
    { 'pymtl_Bits_global_freevar' : pymtl_Bits_global_freevar }
  do_test( a )

def test_pymtl_struct_closure( do_test ):
  a = CaseStructClosureGlobal.DUT()
  a.elaborate()
  a._rtlir_freevar_ref = { 'foo' : a._foo }
  do_test( a )



#=========================================================================
# RTLIRTranslator_L2_cases_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 23, 2019
"""Test the RTLIR translator."""

import pytest

from pymtl3.passes.rtlir.util.test_utility import get_parameter

from ..behavioral.test.BehavioralTranslatorL2_test import test_generic_behavioral_L2
from ..structural.test.StructuralTranslatorL2_test import test_generic_structural_L2
from .TestRTLIRTranslator import TestRTLIRTranslator


def run_test( case, m ):
  if not m._dsl.constructed:
    m.elaborate()
  tr = TestRTLIRTranslator(m)
  tr.translate( m )
  src = tr.hierarchy.src
  assert src == case.REF_SRC

@pytest.mark.parametrize(
  'case', get_parameter('case', test_generic_behavioral_L2) + \
          get_parameter('case', test_generic_structural_L2)
)
def test_generic_L2( case ):
  run_test( case, case.DUT() )



#=========================================================================
# StructuralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L3 structural RTLIR generation pass."""

from .StructuralRTLIRGenL2Pass import StructuralRTLIRGenL2Pass
from .StructuralRTLIRSignalExpr import CurComp


class StructuralRTLIRGenL3Pass( StructuralRTLIRGenL2Pass ):
  pass



#=========================================================================
# BehavioralRTLIRGenL2Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L2 behavioral RTLIR generation pass."""

import ast

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRGenL1Pass import BehavioralRTLIRGeneratorL1


class BehavioralRTLIRGenL2Pass( BasePass ):
  def __call__( s, m ):
    """ generate RTLIR for all upblks of m"""
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()

    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL2( m )
    upblks = {
      'CombUpblk' : get_ordered_upblks(m),
      'SeqUpblk'  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        upblk_info = m.get_update_block_info( blk )
        upblk = visitor.enter( blk, upblk_info[-1] )
        upblk.is_lambda = upblk_info[0]
        upblk.src       = upblk_info[1]
        upblk.lino      = upblk_info[2]
        upblk.filename  = upblk_info[3]
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = upblk

class BehavioralRTLIRGeneratorL2( BehavioralRTLIRGeneratorL1 ):
  def __init__( s, component ):
    super().__init__( component )
    s.loop_var_env = set()
    s.tmp_var_env = set()

    # opmap maps an ast operator to its RTLIR counterpart.
    s.opmap = {
      # Bool operators
      ast.And    : bir.And(),       ast.Or     : bir.Or(),
      # Unary operators
      ast.Invert : bir.Invert(),    ast.Not    : bir.Not(),
      ast.UAdd   : bir.UAdd(),      ast.USub   : bir.USub(),
      # Binary operators
      ast.Add    : bir.Add(),       ast.Sub    : bir.Sub(),
      ast.Mult   : bir.Mult(),      ast.Div    : bir.Div(),
      ast.Mod    : bir.Mod(),       ast.Pow    : bir.Pow(),
      ast.LShift : bir.ShiftLeft(), ast.RShift : bir.ShiftRightLogic(),
      ast.BitOr  : bir.BitOr(),     ast.BitAnd : bir.BitAnd(),
      ast.BitXor : bir.BitXor(),
      # Compare bir.bir.operators
      ast.Eq     : bir.Eq(),        ast.NotEq  : bir.NotEq(),
      ast.Lt     : bir.Lt(),        ast.LtE    : bir.LtE(),
      ast.Gt     : bir.Gt(),        ast.GtE    : bir.GtE()
    }

  def visit_Call( s, node ):
    obj = s.get_call_obj( node )
    # At L2 we add bool type but we do not support instantiating a bool
    # value -- that should always be the result of a comparison!
    if obj is rdt.Bool:
      raise PyMTLSyntaxError( s.blk, node,
        'bool values cannot be instantiated explicitly!' )
    return super().visit_Call( node )

  def visit_Name( s, node ):
    # temporary variable
    if (not node.id in s.closure) and (not node.id in s.globals):
      # check if is a LoopVar or not
      if node.id in s.loop_var_env:
        ret = bir.LoopVar( node.id )
      elif node.id in s.tmp_var_env:
        ret = bir.TmpVar( node.id, s._upblk_name )
      elif isinstance( node.ctx, ast.Load ):
        # trying to load an unregistered tmpvar
        raise PyMTLSyntaxError( s.blk, node,
          'tmpvar ' + node.id + ' used before assignment!' )
      else:
        # This is the first time we see this tmp var
        s.tmp_var_env.add( node.id )
        ret = bir.TmpVar( node.id, s._upblk_name )
      ret.ast = node
      return ret

    else:
      return super().visit_Name( node )

  def visit_If( s, node ):
    cond = s.visit( node.test )
    body = []
    for body_stmt in node.body:
      body.append( s.visit( body_stmt ) )
    orelse = []
    for orelse_stmt in node.orelse:
      orelse.append( s.visit( orelse_stmt ) )
    ret = bir.If( cond, body, orelse )
    ret.ast = node
    return ret

  def visit_For( s, node ):
    # First fill the loop_var, start, end, step fields
    if node.orelse != []:
      raise PyMTLSyntaxError( s.blk, node,
        "for loops cannot have 'else' branch!" )
    if not isinstance( node.target, ast.Name ):
      raise PyMTLSyntaxError( s.blk, node,
        "The loop index must be a temporary variable!" )
    loop_var_name = node.target.id

    # Check whether loop_var_name has been defined before
    if loop_var_name in s.loop_var_env:
      raise PyMTLSyntaxError( s.blk, node,
        "Redefinition of loop index " + loop_var_name + "!" )

    # Add loop_var to the loop variable environment
    s.loop_var_env.add( loop_var_name )
    var = bir.LoopVarDecl( node.target.id )
    if not isinstance( node.iter, ast.Call ):
      raise PyMTLSyntaxError( s.blk, node,
        "for loops can only use range() after 'in'!" )
    if node.iter.func.id != 'range':
      raise PyMTLSyntaxError( s.blk, node,
        "for loops can only use range() after 'in'!" )
    args = node.iter.args

    if len( args ) == 1:
      # range( end )
      start = bir.Number( 0 )
      end = s.visit( args[0] )
      step = bir.Number( 1 )
    elif len( args ) == 2:
      # range( start, end )
      start = s.visit( args[0] )
      end = s.visit( args[1] )
      step = bir.Number( 1 )
    elif len( args ) == 3:
      # range( start, end, step )
      start = s.visit( args[0] )
      end = s.visit( args[1] )
      step = s.visit( args[2] )
    else:
      raise PyMTLSyntaxError( s.blk, node,
        "1~3 arguments should be given to range!" )

    # Then visit all statements inside the loop
    body = []
    for body_stmt in node.body:
      body.append( s.visit( body_stmt ) )

    # Before we return, clear the loop variable in the loop variable
    # environment
    s.loop_var_env.remove( loop_var_name )
    ret = bir.For( var, start, end, step, body )
    ret.ast = node
    return ret

  def visit_BoolOp( s, node ):
    if not type(node.op) in s.opmap:
      raise PyMTLSyntaxError( s.blk, node,
        'Operator ' + str( node.op ) + ' is not supported!' )
    op  = s.opmap[ type( node.op ) ]
    op.ast = node.op

    values = []
    for value in node.values:
      values.append( s.visit( value ) )

    ret = bir.BoolOp( op, values )
    ret.ast = node
    return ret

  def visit_BinOp( s, node ):
    left  = s.visit( node.left )
    right = s.visit( node.right )
    if not type(node.op) in s.opmap:
      raise PyMTLSyntaxError( s.blk, node,
        'Operator ' + str( node.op ) + ' is not supported!' )
    op  = s.opmap[ type( node.op ) ]
    op.ast = node.op
    ret = bir.BinOp( left, op, right )
    ret.ast = node
    return ret

  def visit_UnaryOp( s, node ):
    if not type(node.op) in s.opmap:
      raise PyMTLSyntaxError( s.blk, node,
        'Operator ' + str( node.op ) + ' is not supported!' )
    op  = s.opmap[ type( node.op ) ]
    op.ast = node.op
    operand = s.visit( node.operand )
    ret = bir.UnaryOp( op, operand )
    ret.ast = node
    return ret

  def visit_IfExp( s, node ):
    cond = s.visit( node.test )
    body = s.visit( node.body )
    orelse = s.visit( node.orelse )
    ret = bir.IfExp( cond, body, orelse )
    ret.ast = node
    return ret

  def visit_Compare( s, node ):
    if not type(node.ops[0]) in s.opmap:
      raise PyMTLSyntaxError( s.blk, node,
        'Operator ' + str( node.ops[0] ) + ' is not supported!' )
    if len( node.ops ) != 1 or len( node.comparators ) != 1:
      raise PyMTLSyntaxError( s.blk, node,
        'Comparison can only have 2 operands!' )

    op  = s.opmap[ type( node.ops[0] ) ]
    op.ast = node.ops[0]
    left = s.visit( node.left )
    right = s.visit( node.comparators[0] )
    ret = bir.Compare( left, op, right )
    ret.ast = node
    return ret



#=========================================================================
# YosysBehavioralTranslatorL4.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog L4 behavioral translator."""

from pymtl3.passes.backends.sverilog.translation.behavioral.SVBehavioralTranslatorL4 import (
    BehavioralRTLIRToSVVisitorL4,
    SVBehavioralTranslatorL4,
)
from pymtl3.passes.rtlir import RTLIRType as rt

from .YosysBehavioralTranslatorL3 import (
    YosysBehavioralRTLIRToSVVisitorL3,
    YosysBehavioralTranslatorL3,
)


class YosysBehavioralTranslatorL4(
    YosysBehavioralTranslatorL3, SVBehavioralTranslatorL4 ):

  def _get_rtlir2sv_visitor( s ):
    return YosysBehavioralRTLIRToSVVisitorL4

class YosysBehavioralRTLIRToSVVisitorL4(
    YosysBehavioralRTLIRToSVVisitorL3, BehavioralRTLIRToSVVisitorL4 ):

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for accessing interface attribute in L4.
    """
    if isinstance( node.value.Type, rt.InterfaceView ):
      value = s.visit( node.value )
      s.signal_expr_prologue( node )
      attr = node.attr
      s.check_res( node, attr )
      node.sexpr['s_attr'] += "__{}"
      node.sexpr['attr'].append( attr )
      return s.signal_expr_epilogue(node, f"{value}.{attr}")

    else:
      return super().visit_Attribute( node )



from .arbiters import RoundRobinArbiter, RoundRobinArbiterEn
from .arithmetics import (
    Adder,
    And,
    Incrementer,
    LEComparator,
    LeftLogicalShifter,
    LTComparator,
    Mux,
    RightLogicalShifter,
    Subtractor,
    ZeroComparator,
)
from .Crossbar import Crossbar
from .queues import BypassQueueRTL, NormalQueueRTL, PipeQueueRTL
from .RegisterFile import RegisterFile
from .registers import Reg, RegEn, RegEnRst, RegRst

#  from .enrdy_queues import BypassQueue1RTL, NormalQueue1RTL, PipeQueue1RTL
#  from .valrdy_queues import (
    #  BypassQueue1RTL,
    #  NormalQueue1RTL,
    #  NormalQueueRTL,
    #  PipeQueue1RTL,
#  )
